/* Séquence qui permet de tester le matériel du simulateur d'ÉNIGMA. Les seuls périphériques à tester
   sont le bruiteur passif piloté par D4 et la LED bleue branchée sur D5.
   Le programme commence par générer le mot BONJOUR en code morse, sur le bruiteur et sur la LED bleue
   puis fait clignoter la LED Arduino en D13 rapidement et à 1Hz la LED bleue.

Compilateur v1.7.9 :                        PGM                 RAM dynamique
Avant modification :                  3128 Octets (10%)      285 Octets (13%)   */

//--------------------- Définition des broches d'E/S ------------------------
#define BUZZER 4
#define LED_bleue 5
#define LED_Arduino 13

//------------------------ Constantes du programme --------------------------
#define Version "19/01/2024" // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Vitesse_USB 57600  // Taux de transfert sur la ligne USB @@@@@@@@@@@@@@@@@@@@@@@@
#define Delai_voulu 500  // Temporisation en mS @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Rapide false  // Cadence de la manipulation morse @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

//------------------------- Variables du programme --------------------------

boolean Morse = true;

byte Caractere; // Contiendra l'ordre alphabétique du caractère en cours de traitement.

byte BONJOUR[7] = {2,15,14,10,15,21,18};

byte CODE_MORSE[27] = {0,66,129,133,97,32,132,99,128,64,142,101,130,67,65,
                       103,134,139,98,96,33,100,129,102,137,141,131};
     // NOTE : Pour simplifier le codage, l'indice 0 ne sera jamais utilisé et on aura
     //        Indice 1 pour A, indide 2 pour B etc.

unsigned long Temps_1_Ecoule_depuis_RESET = 0;
unsigned long Temps_2_Ecoule_depuis_RESET = 0;

//---------------------------------------------------------------------------

void setup() {
  Serial.begin(57600); Serial.println(); Serial.print("Version du ");
  Serial.println(Version); Serial.println();
  pinMode(LED_Arduino, OUTPUT);
  pinMode(LED_bleue, OUTPUT); digitalWrite(LED_bleue,LOW);
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER,LOW);
  Genere_BONJOUR_en_Morse();
  Serial.println(); Serial.println(); Serial.print("Boucle de clignotement infinie.");}
 
void loop() {
  //================= Chronométrage pour le délai écoulé ================
  if (millis() - Temps_1_Ecoule_depuis_RESET > Delai_voulu)  {
     Temps_1_Ecoule_depuis_RESET = millis(); // Réarmer le chronométrage. (TOP 1.)
     //----------- Actions à traiter pour chaque délai désiré. ----------
     Inverser_LED_bleue();}
  //================= Chronométrage pour le délai rapide ================
  if (millis() - Temps_2_Ecoule_depuis_RESET > 50)  {
     Temps_2_Ecoule_depuis_RESET = millis(); // Réarmer le chronométrage. (TOP 2.)
     //----------- Actions à traiter pour chaque délai écoulé. ----------
     Inverser_LED_Arduino();}}

/* NOTE : La technique dans la boucle de base pour déclencher des actions à cadences précises
          consiste à utiliser les ressources de l'ATmega328 de façon à ne pas se servir de
          la fonction delay(nnn) qui bloque le processeur durant l'attente. */

//======================== Procédures de servitude =============================

void Inverser_LED_Arduino() {digitalWrite(LED_Arduino,!digitalRead(LED_Arduino));}

void Inverser_LED_bleue() {digitalWrite(LED_bleue,!digitalRead(LED_bleue));}

void POINT() {// Génère automatiquement le silence de séparation des éléments.
  if (Morse) {
      int T_element = 84; if (Rapide) T_element = 57;
      digitalWrite(LED_bleue,HIGH); digitalWrite(BUZZER,HIGH); delay(T_element);
      digitalWrite(LED_bleue,LOW); digitalWrite(BUZZER,LOW); delay(T_element);}
  Serial.print('.');}
  
void TRAIT() {// Génère automatiquement le silence de séparation des éléments.
  if (Morse) {
      int T_element = 252; if (Rapide) T_element = 171;
      digitalWrite(LED_bleue,HIGH); digitalWrite(BUZZER,HIGH); delay(T_element);
      digitalWrite(LED_bleue,LOW); digitalWrite(BUZZER,LOW); delay(T_element);}
  Serial.print('-');}

void TRANSCODE() {// Génère en Morse le Caractère représenté par l'ordre de la lettre dans l'alphabet.
 byte Lettre, Nb_Elements;
 int T_element = 252; if (Rapide) T_element = 171;
 Lettre = CODE_MORSE[Caractere]; Serial.print(char(64 + Caractere)); Serial.print("  ");
 // Ci-dessus : Pour afficher les caractères il suffit d'ajouter 64 car le code de A = 65.
 Nb_Elements = Lettre >> 5; // Isole le nombre d'éléments par cinq décalages à droite.
 for (Nb_Elements; Nb_Elements > 0; Nb_Elements--) {
   if ((Lettre & 1) > 0) TRAIT(); else POINT(); Lettre = Lettre >> 1;}
 delay (T_element); Serial.println();} // Silence de 0,4S avec le délai entre éléments.

void Genere_BONJOUR_en_Morse() {
  for (byte Num_caractere = 0; Num_caractere < 7; Num_caractere++) {
       Caractere = BONJOUR[Num_caractere]; TRANSCODE();}}
       
  

