
/* Séquence qui permet D'ajouter l'émulation du Rotor I au dialogue entre l'opérateur et le
   simulateur d'ÉNIGMA. Pour utiliser ce programme il suffit de proposer des caractères en
   mode LETTRES. Ils "traversent" le Rotor dans les deux sens, donc au final on retrouve le
   caractère d'entrée. Entre '<' et '>' est affiché le caractère "au niveau du Réflecteur.
   
            ******************************************************************
            * Ce démonstrateur constitue le noyau du programme d'utilisation *
            *       et sera complété au fur et à mesure du développement.    *
            ******************************************************************
   
Compilateur v1.7.9 :                        PGM                 RAM dynamique
Avant modification :                  6878 Octets (22%)        1531 Octets (74%)
Apes ajout du Rotor I :               7058 Octets (22%)        1619 Octets (74%)   */

//--------------------- Définition des broches d'E/S ------------------------
#define BUZZER 4
#define LED_bleue 5
#define LED_Arduino 13

//------------------------ Constantes du programme --------------------------
#define Version "25/01/2024" // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Vitesse_USB 57600  // Taux de transfert sur la ligne USB @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Taille 60 // Longueur maximale de la chaine de dialogue @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// Taille réservée pour une chaine de 60 plus le délimiteur.
// Inutile de réserver plus de 60 caractère car la limite du FIFO USB est de 63 caractères.

//------------------------- Variables du programme --------------------------
     // Rapide : Cadence de la manipulation Morse.
     // Texte : Mode TEXTE au lieu du mode LETTRES.
boolean Mode_COMMANDE, Morse, Rapide, Texte;

     // Caractere contiendra l'ordre alphabétique du caractère en cours de traitement.
     // Index : Pointeur pour le cryptage dans la chaine TEXTE.
byte Caractere, Compteur_pour_separer, Compteur_pour_ligne, ERREUR, Index, LGR_chaine, L_autorisee;

byte CODE_MORSE[27] = {0,66,129,133,97,32,132,99,128,64,142,101,130,67,65,
                       103,134,139,98,96,33,100,129,102,137,141,131};
     // NOTE : Pour simplifier le codage, l'indice 0 ne sera jamais utilisé et on aura
     //        Indice 1 pour A, indide 2 pour B etc.
     
byte Rotor_de_DroiteD[26] = {5,11,13,6,12,7,4,17,22,26,14,20,15,23,25,8,24,21,19,16,1,9,2,18,3,10};
byte Rotor_de_DroiteR[26] = {21,23,25,7,1,4,6,16,22,26,2,5,3,11,13,20,8,24,19,12,18,9,14,17,15,10};

unsigned long TOP_chronometre;

char Chaine_Memorisee[Taille+2]; // Tampon mémoire pour la chaine entrante. taille + 2 car il faut un
                          // emplacement pour le délimiteur et un emplacement pour "pouvoir déborder".

//-------------------------------------------------------------------------------------

void setup() {
  Serial.begin(57600); Serial.println(); Serial.print("Version du ");
  Serial.println(Version); Serial.println();
  pinMode(LED_Arduino, OUTPUT); pinMode(LED_bleue, OUTPUT); digitalWrite(LED_bleue,LOW);
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER,LOW);
  //--------------------------- Initialisation des variable ---------------------------
  L_autorisee = 1; Mode_COMMANDE = true; Morse = false; Rapide = false; Texte = false;
  TOP_chronometre = 0;
  Affiche_le_MENU();}
 
void loop() {
  Attendre_une_chaine_sur_RX();
  if (Mode_COMMANDE) Traiter_la_COMMANDE(); else Traiter_le_CRYPTAGE();}

//======================== Procédures de servitude =============================

void BIP() {digitalWrite(BUZZER,HIGH); delay(150); digitalWrite(BUZZER,LOW);}

void N_ESPACES(byte Nb) {for(byte N = 0; N < Nb; N++)  Serial.print(' ');}

void Aff_46_etoiles() {for(byte N = 0; N < 46; N++) Serial.print('*');}

void Separer() {Aff_46_etoiles(); Serial.println(); Serial.print('*');}

void Terminer(byte Nb) {
  for(byte N = 0; N < Nb; N++) Serial.print(' '); Serial.println('*'); Serial.print('*');}

void Affiche_caractere_Num() {Serial.print(char(Caractere + 64));}

void Aff_OUI_NON(boolean OUI) {if (OUI) Serial.print("OUI"); else Serial.print("NON");}

void Afficher_la_Commande() {
  if (Mode_COMMANDE) Serial.print(" Commande -> ");
  else if (!Texte) Serial.print(" Une lettre -> ");}

void Affiche_le_MENU() {
  Serial.print('*'); Separer(); N_ESPACES(13); Serial.print("Menu des COMMANDES."); Terminer(13);
  Separer(); Serial.print(" ? ou , : Rappel des commandes."); Terminer(14);
  Serial.print(" I ou i : Mode INITIALISATION."); Terminer(15);
  Serial.print(" S ou s : Sauvegarder la configuration."); Terminer(6);
  Serial.print(" R ou r : Restituer la configuration."); Terminer(8);
  Serial.print(" M ou m : Morse sonore OUI / NON."); Terminer(12);
  Serial.print(" V ou v : Vitesse du debit du code Morse."); Terminer(4);
  Serial.print(" T ou t : Mode TEXTE."); Terminer(24);
  Serial.print(" L ou l : Mode LETTRES."); Terminer(22);
  Serial.print(" C ou c : Affiche la CONFIGURATION."); Terminer(10);
  Serial.print(" P ou p : Espace disponible sous la PILE."); Terminer(4);
  Serial.print(" & ou 1 : Passer en Mode CRYPTAGE."); Terminer(11);
  Separer(); N_ESPACES(13); Serial.print("Menu INITIALISATION."); Terminer(12);
  Separer(); Serial.print(" R ou r : Choix du ROTOR et de sa position ."); Terminer(1);
  Serial.print(" B ou b et C ou c : Choix du REFLECTEUR."); Terminer(5);
  Serial.print(" F ou f : Configurer les FICHES croisees."); Terminer(4);
  Serial.print(" G ou g : Groupes d'identification."); Terminer(10);
  Serial.print(" E ou e suivi de R, C, F ou G : Corriger."); Terminer(4);
  Serial.print("   (C pour Reflecteur de Complement.)"); Terminer(8);
  Separer(); N_ESPACES(9); Serial.print("CRYPTAGE en mode LETTRES."); Terminer(11);
  Separer(); Serial.print(" & ou 1 : Passer en Mode COMMANDES."); Terminer(10);
  Serial.print(" Frapper Uniquement des lettres et valider."); Terminer(2);
  Separer(); N_ESPACES(10); Serial.print("CRYPTAGE en mode TEXTE."); Terminer(12);
  Separer(); Serial.print(" & ou 1 : Passer en Mode COMMANDES."); Terminer(10);
  Serial.print("     Frapper des mots. (ESPACE valide.)"); Terminer(6); Aff_46_etoiles();
  Serial.println(); Serial.println();}

void POINT() {// Génère automatiquement le silence de séparation des éléments.
  if (Morse) {
      int T_element = 84; if (Rapide) T_element = 57;
      digitalWrite(LED_bleue,HIGH); digitalWrite(BUZZER,HIGH); delay(T_element);
      digitalWrite(LED_bleue,LOW); digitalWrite(BUZZER,LOW); delay(T_element);}
  Serial.print('.');}
  
void TRAIT() {// Génère automatiquement le silence de séparation des éléments.
  if (Morse) {
      int T_element = 252; if (Rapide) T_element = 171;
      digitalWrite(LED_bleue,HIGH); digitalWrite(BUZZER,HIGH); delay(T_element);
      digitalWrite(LED_bleue,LOW); digitalWrite(BUZZER,LOW); delay(T_element);}
  Serial.print('-');}

void TRANSCODE() {// Génère en Morse le Caractère représenté par l'ordre de la lettre dans l'alphabet.
 byte Lettre, Nb_Elements;
 int T_element = 252; if (Rapide) T_element = 171;
 Lettre = CODE_MORSE[Caractere]; Serial.print("  ");
 Nb_Elements = Lettre >> 5; // Isole le nombre d'éléments par cinq décalages à droite.
 for (Nb_Elements; Nb_Elements > 0; Nb_Elements--) {
   if ((Lettre & 1) > 0) TRAIT(); else POINT(); Lettre = Lettre >> 1;}
 delay (T_element); Serial.println();} // Silence de 0,4S avec le délai entre éléments.

//====================================================================================================

void Faciliter_la_manipulation() {
  switch(Caractere) {
    case ',' : {Caractere = '?'; break;} case '1' : {Caractere = '&'; break;}
    case 224 : {Caractere = 'a'; break;} case 226 : {Caractere = 'a'; break;}
    case 232 : {Caractere = 'e'; break;} case 233 : {Caractere = 'e'; break;}
    case 234 : {Caractere = 'e'; break;} case 235 : {Caractere = 'e'; break;}
    case 238 : {Caractere = 'i'; break;} case 239 : {Caractere = 'i'; break;}
    case 244 : {Caractere = 'o'; break;} case 231 : {Caractere = 'c'; break;}
    case 251 : {Caractere = 'u'; break;} case 252 : {Caractere = 'u'; break;}
    case '.' : Caractere = ' ';}} // Le '.' sera ignoré comme l'espace.

void Attendre_une_chaine_sur_RX() {
  byte I = 0; LGR_chaine = 0; // Pointeur dans Chaine_Memorisee.
  Afficher_la_Commande();
  while(!Serial.available()) { // Attendre un premier caractere.
     //================= Chronométrage pour le délai écoulé ================
     if (millis() - TOP_chronometre > 50)  {
        TOP_chronometre = millis(); digitalWrite(LED_Arduino,!digitalRead(LED_Arduino));}}
     //---------------------------------------------------------------------
  while(Serial.available()) {
     Caractere = Serial.read();
 //    if(Caractere == ',') Caractere = '?'; if(Caractere == '1') Caractere = '&';
     Faciliter_la_manipulation();
     if ((Caractere > 96) && (Caractere < 123)) Caractere = Caractere & 223; // Bien moins couteux
              // en octets que Chaine_Memorisee.toUpperCase() pour forcer la chaine en MAJUSCULES.
     if (LGR_chaine < (Taille + 1)) {
        Chaine_Memorisee[LGR_chaine] = Caractere; LGR_chaine++; delay(5);}} // Délai impératif.
  Chaine_Memorisee[LGR_chaine] = '\0'; // Placer le délimiteur de fin de texte.
  //----------------------------------------------------------------------------------------
  if (Mode_COMMANDE) {Serial.print('[');
     while(Chaine_Memorisee[I] != '\0') {Serial.print(char(Chaine_Memorisee[I])); I++;}
     Serial.print(']'); Serial.println(); digitalWrite(LED_Arduino, LOW);}}
  
void Mode_INITIALISATION() {Serial.println(" Traite l'INITIALISATION.");}
void Sauvegarde_Initialisation() {Serial.println(" Traite la SAUVEGARDE.");}
void Restitue_Initialisation() {Serial.println(" Traite le RECHARGEMENT.");}
void Affiche_la_CONFIGURATION() {Serial.println(" Affiche la CONFIGURATION.");}

int SRAM_LIBRE() { // Fonction qui retourne la taille de SRAM disponible.
  extern int __heap_start, *__brkval; // Déclaration des deux pointeurs dédiés.
  byte BIDON; // Dernière variable allouée, donc occupe le "haut" du TAS.
  if (__brkval == 0) {return (int) &BIDON -(int) &__heap_start;}
  else {return (int) &BIDON -(int) __brkval;}}

void Affiche_espace_sous_la_PILE() {
  Serial.print(" RAM DYNAMIQUE (PILE - TAS)  =  "); Serial.print (SRAM_LIBRE());
  Serial.println(" octets.");}

//================================== Traiter un CRYPTAGE ==================================

void Crypte_le_caractere() {// Reçoit l'ordre alphabétique du caractère crypté.
  byte I;
  //--------------------- Brouille_le_caractere ---------------------
  Caractere = Rotor_de_DroiteD[Caractere-1]; // Traverse le Rotor dans le sens Direct.
  Serial.print('<'); Affiche_caractere_Num(); Serial.print("> "); //I = 0;
  Caractere = Rotor_de_DroiteR[Caractere-1];} // Tracerse en sens Réfléchi.
  
void Aff_texte_avant_erreur() {
     if (Texte) {Serial.println(); Serial.print('<');
         for(byte I = 0; I < Index; I++) Serial.print(Chaine_Memorisee[I]); Serial.print("> ");}}

void Aff_ERREUR(byte Num_ERR) {
  switch (Num_ERR) {
     case 1 : {Serial.print(" Attention : LGR > MAX."); Index = 60; Aff_texte_avant_erreur();
               Index = 0; L_autorisee = 255; // Pour éviter autant de BIPs que de dépassement.
               Compteur_pour_separer = 255; Compteur_pour_ligne = 255; break;}
     case 2 : {Serial.print(" Caractere de COMMANDE invalide."); break;}
     case 3 : {if (Texte) {
                   Aff_texte_avant_erreur(); Index = LGR_chaine;} // Ne pas envoyer la suite.    
               Serial.print('['); Serial.print(char(Caractere));
               Serial.print(']'); Serial.print(" Caractere de CRYPTAGE invalide.");}}
  Serial.println(); BIP();}

  
void Verification_du_cryptage() {
  ERREUR = 0; Caractere = Chaine_Memorisee[Index];
  if (LGR_chaine > L_autorisee) ERREUR = 1;
  if ((Caractere < 'A') || (Caractere > 'Z')) ERREUR = 3;
  if ((Caractere == ' ') && (!Texte)) ERREUR = 3;
  if (ERREUR > 0) Aff_ERREUR(ERREUR);}

void Aff_retour_Commande() {Serial.println(); Serial.println(" Retour au mode COMMANDE.");}

void Traiter_le_Caractere() {
  if (Caractere == '&') Mode_COMMANDE = true;
  if (Mode_COMMANDE) Aff_retour_Commande();
  else {Verification_du_cryptage();
       if (ERREUR == 0) {
          Caractere = (byte(Chaine_Memorisee[Index])) - 64; Crypte_le_caractere();
          Affiche_caractere_Num();}}}

void Traiter_le_TEXTE() {// Index = 0 en entrée de cette procédure.
  for(Index; Index < LGR_chaine; Index++) {
      Caractere = Chaine_Memorisee[Index];
      if (!(Caractere == ' ')) {  // Ignorer ' ' en mode TEXTE.
         Traiter_le_Caractere(); Compteur_pour_separer++; Compteur_pour_ligne++;
      if (Compteur_pour_separer == 6) {Serial.print(' '); Compteur_pour_separer = 0;}
      if (Compteur_pour_ligne == 54) {Serial.println(); Compteur_pour_ligne = 0;}}}}

void Traiter_le_CRYPTAGE() { Index = 0; // Impératif pour Verification_du_cryptage().
  if (Texte) {L_autorisee = Taille; Traiter_le_TEXTE();}
  else {Caractere = Chaine_Memorisee[0]; L_autorisee = 1;
        if (Caractere != '&') Verification_du_cryptage();
        if (ERREUR == 0) {
           if (Caractere == '&') {
               Mode_COMMANDE = true; Serial.println(); Aff_retour_Commande();}
           else {Serial.print('['); Serial.print(char(Caractere)); Serial.print("] devient [");
                 Traiter_le_Caractere(); Serial.print(']'); TRANSCODE();}}
        else  ERREUR = 0;}} // Impératif pour reprendre la main avec '&'.

//================================= Traiter une COMMANDE ==================================

void Analyseur_syntaxique() { // Refuse les caractères de commande invalides.
  ERREUR = 0; L_autorisee = 1; Caractere = Chaine_Memorisee[0];
  if (LGR_chaine > L_autorisee) ERREUR = 1; 
  switch (Caractere) {
      case '?' : {; break;} case 'I' : {; break;} case 'S' : {; break;} case 'R' : {; break;}
      case 'M' : {; break;} case 'V' : {; break;} case 'T' : {; break;} case 'L' : {; break;}
      case 'C' : {; break;} case 'P' : {; break;} case '&' : {; break;} default : ERREUR = 2;}
  if (ERREUR > 0) Aff_ERREUR(ERREUR);}

void Affiche_le_mode(byte Mode) {
  Serial.print(" Passage en Mode ");
  switch (Mode) {
    case 1 : {if (Mode_COMMANDE) {Serial.println("Commande.");}
              else Serial.println("CRYPTAGE."); break;}
    case 2 : {Serial.println("TEXTE. (60 caracteres MAXI.)"); break;}
    case 3 : Serial.println("LETTRES.");}}

void Traiter_la_COMMANDE() {
  byte I = 0; // Pointer le début de la chaine TEXTE.
  Analyseur_syntaxique(); // Vérifie les COMMANDES.
  if (ERREUR == 0)
     switch (Caractere) {
       case '?' : {Affiche_le_MENU(); break;}
       case 'I' : {Mode_INITIALISATION(); break;}
       case 'S' : {Sauvegarde_Initialisation(); break;}
       case 'R' : {Restitue_Initialisation(); break;}
       case 'M' : {Morse = !Morse; Serial.print(" Morse sonore = "); Aff_OUI_NON(Morse);
                   Serial.println(); break;}
       case 'V' : {Rapide = !Rapide; Serial.print(" Vitesse Morse = ");
                   if (Rapide) Serial.println("Rapide."); else Serial.println("Lente."); break;}
       case 'T' : {Texte = true; Affiche_le_mode(2); break;}
       case 'L' : {Texte = false; Affiche_le_mode(3); break;}
       case 'C' : {Affiche_la_CONFIGURATION(); break;}
       case 'P' : {Affiche_espace_sous_la_PILE(); break;}
       case '&' : {Mode_COMMANDE = false; Compteur_pour_separer = 0; Compteur_pour_ligne = 0;
                   Serial.println(); Serial.println(" Passage au mode CRYPTAGE.");}}}
  
 

       
  

