
/* Démonstrateur qui met en place ENTIEREMENT la procédure d'initialisation de la machine.
   Émulant le MENU simplifié il utilise les nouveaux textes EEPROM de l'outil P00B.
   Maintenant si le Morse sonore est suspendu les BIPs d'ERREUR aussi. (Suspendus par défaut.)
   
   UTILISATION : Commande 'i' pour entrer dans le menu "linéaire" d'initialisation.
                 Commande 'c' pour faire afficher l'initialisation actuelle.
                 Initialise les Rotors, le Réflecteur, les FICHES croisées ainsi que le
                 groupe d'identification.
                 >>>>>> Manipulations proposées dans le didacticiel en page 30.
   
            ******************************************************************
            * Ce démonstrateur constitue le noyau du programme d'utilisation *
            * et sera complété au fur et à mesure du développement.          *
            ******************************************************************

Compilateur v1.7.9 :                                    PGM                   RAM dynamique
P05 avant toute modification :                    7546 Octets (24%)        601 Octets (29%)
Passage au MENU simplifié :                       7358 Octets (23%)        575 Octets (28%)
Ajout initialisation Fiches et Groupe ID :        8306 Octets (27%)        598 Octets (29%)
BIP d'erreur silencieux si NON Morse :            8290 Octets (26%)        590 Octets (28%) */

//--------------------- Définition des broches d'E/S ------------------------
#define BUZZER 4
#define LED_bleue 5

//------------------------ Constantes du programme --------------------------
#define Version "05/02/2024" // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Vitesse_USB 57600  // Taux de transfert sur la ligne USB @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Taille 60 // Longueur maximale de la chaine de dialogue @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// Taille réservée pour une chaine de 60 plus le délimiteur.
// Inutile de réserver plus de 60 caractère car la limite du FIFO USB est de 63 caractères.

//------------------------- Variables du programme --------------------------

     // Rapide : Cadence de la manipulation Morse.
     // Texte : Mode TEXTE au lieu du mode LETTRES.
boolean Mode_COMMANDE, Mode_Initialisation, Morse, OUI, Rapide, Texte;

     // Caractere contiendra l'ordre alphabétique du caractère en cours de traitement.
     // Index : Pointeur pour le cryptage dans la chaine TEXTE.
byte Caractere, Compteur_pour_separer, Compteur_pour_ligne, ERREUR, Index, Index_de_Droite,
     Index_de_Gauche, Index_du_Centre, LGR_chaine, L_autorisee, Num_ERR, Rotor_de_Droite, Rotor_de_Gauche,
     Rotor_du_Centre;

char Orientation_de_Droite, Orientation_du_Centre,Orientation_de_Gauche, REFLECTEUR ;

int Nombre, PTR;

byte CODE_MORSE[27] = {0,66,129,133,97,32,132,99,128,64,142,101,130,67,65,
                       103,134,139,98,96,33,100,129,102,137,141,131};
     // NOTE : Pour simplifier le codage, l'indice 0 ne sera jamais utilisé et on aura
     //        Indice 1 pour A, indide 2 pour B etc.

char GROUPE_de_tete[12] = {};
char FICHES_croisees[20] = {};
boolean Lettre_utilisee[26];

unsigned long TOP_chronometre;

char Chaine_Memorisee[Taille+2]; // Tampon mémoire pour la chaine entrante. taille + 2 car il faut un
                          // emplacement pour le délimiteur et un emplacement pour "pouvoir déborder".

//-------------------------------------------------------------------------------------

void setup() {
  Serial.begin(57600); Serial.println(); Aff_TEXTE_EEPROM(439,12); //"Version du "
  Serial.println(Version);
  pinMode(LED_bleue, OUTPUT); digitalWrite(LED_bleue,LOW); pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER,LOW);
  //--------------------------- Initialisation des variable ---------------------------
  L_autorisee = 1; Mode_COMMANDE = true; Mode_Initialisation = false, Morse = false; Rapide = false;
  Texte = false; TOP_chronometre = 0;
  Affiche_le_MENU();}
 
void loop() {
  Attendre_une_chaine_sur_RX();
  if (Mode_COMMANDE) Traiter_la_COMMANDE(); else Traiter_le_CRYPTAGE();}

//==================== Procédures Écriture / Lecture EEPROM ====================
byte Lire_Octet_EEPROM(int Pointeur) {
  return eeprom_read_byte((unsigned char *) Pointeur);}

byte Ecrire_Octet_EEPROM(int Pointeur, byte OCTET) {
  eeprom_write_byte((unsigned char*) Pointeur,OCTET); PTR++;}

void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {
    for (byte I=0; I<Longueur; I++) {Serial.print(char(Lire_Octet_EEPROM(PTR))); PTR++;}}

//======================== Procédures de servitude =============================

void BIP() {if (Morse) {digitalWrite(BUZZER,HIGH); delay(150); digitalWrite(BUZZER,LOW);}}

void N_ESPACES(byte Nb) {for(byte N = 0; N < Nb; N++)  Serial.print(' ');}
void Aff_etoile() {Serial.print('*');}
void Separer() {Aff_46_etoiles(); Serial.println(); N_ESPACES(1); Aff_etoile();}
void Aff_46_etoiles() {for(byte N = 0; N < 46; N++) Aff_etoile();}
void Terminer(byte Nb) {
  for(byte N = 0; N < Nb; N++) N_ESPACES(1); Serial.println('*'); N_ESPACES(1); Serial.print('*');}
void Affiche_caractere_Num() {Serial.print(char(Caractere + 64));}
void AFF_crochet_ouvrant() {Serial.print('[');}
void AFF_crochet_fermant() {Serial.print(']');}
void Affiche_triple_fleche_ouvrantes() {Serial.print(" >>> ");}
void Affiche_triple_fleche_fermantes() {Serial.print(" <<<");}
void Affiche_Egal() {Serial.print(" = ");}
void Fleche() {Aff_TEXTE_EEPROM(42,4);}
void AFF_LF_Espace() {Serial.println(); N_ESPACES(1);}
void Aff_OUI_NON(boolean OUI) {if (OUI) Aff_TEXTE_EEPROM(187,3); else Aff_TEXTE_EEPROM(193,3);}
void Aff_texte_Caractere() {Aff_TEXTE_EEPROM(598,11);} // " Caractère "
void Aff_texte_invalide() {N_ESPACES(1); Aff_TEXTE_EEPROM(610,9);} // " invalide."
void Affiche_passage_au_mode() {Aff_TEXTE_EEPROM(630,17);}
void Aff_CRYPTAGE() {Aff_TEXTE_EEPROM(375,9);}
void Aff_de_Droite() {Aff_TEXTE_EEPROM(410,10);} 
void Aff_du_Centre() {Aff_TEXTE_EEPROM(430,10);}
void Aff_de_Gauche() {Aff_TEXTE_EEPROM(420,10);}
void Aff_Index() {N_ESPACES(1); Aff_TEXTE_EEPROM(656,5); N_ESPACES(1);}
void Aff_ROTOR() { N_ESPACES(1); Aff_TEXTE_EEPROM(462,5); N_ESPACES(1);}
void Aff_Rotor_de_droite() {Aff_ROTOR(); Aff_de_Droite();}
void Aff_Rotor_du_centre() {Aff_ROTOR(); Aff_du_Centre();}
void Aff_Rotor_de_gauche() {Aff_ROTOR(); Aff_de_Gauche();}
void Aff_Position_de_droite() {Aff_ROTOR(); Aff_de_Droite();}
void Aff_Position_du_centre() {Aff_ROTOR(); Aff_du_Centre();}
void Aff_Position_de_gauche() {Aff_ROTOR(); Aff_de_Gauche();}
void Aff_Orientation() { N_ESPACES(1); Aff_TEXTE_EEPROM(451,11);}
void Aff_de_puis_Espace() {Aff_TEXTE_EEPROM(420,3);}
void Aff_Fiche_num() {Aff_TEXTE_EEPROM(515,6); N_ESPACES(1); Aff_TEXTE_EEPROM(654,2);} // " FICHE nº"
void Aff_Groupe_Identification() {N_ESPACES(1); Aff_TEXTE_EEPROM(384,23);} // "GROUPE d'identification"

void Aff_retour_Commande() {Serial.println(); Aff_TEXTE_EEPROM(646,8); Aff_TEXTE_EEPROM(639,8);
     Aff_TEXTE_EEPROM(22,10); Serial.println();} // " Retour au mode COMMANDES."

void Afficher_la_Commande() {
  if (Mode_COMMANDE) {Aff_TEXTE_EEPROM(21,9); Fleche();} // " COMMANDE -> "
  else {if (!Texte) {N_ESPACES(1); Aff_TEXTE_EEPROM(32,14);}}} // " Une lettre -> ")

void Affiche_le_caractere_entre_crochets() {
  AFF_crochet_ouvrant(); Serial.print(char(Caractere)); AFF_crochet_fermant();}

void Confirmer() {
  Serial.print(" Confirmer (O/N)"); Fleche(); Attendre_une_Reponse_sur_RX(); OUI = Caractere == 'O';}

void Affiche_le_MENU() {
   Serial.println(); N_ESPACES(1); Aff_etoile(); Separer();
  N_ESPACES(13); Aff_TEXTE_EEPROM(13,19); // "Menu des COMMANDES."
  Terminer(13); Separer(); Aff_TEXTE_EEPROM(45,31); // " ? ou , : Rappel des commandes."
  Terminer(14); Aff_TEXTE_EEPROM(76,30); Terminer(15); // " I ou i : Mode INITIALISATION."
  Aff_TEXTE_EEPROM(532,36);  Terminer(9); // " E ou e suivi de R, F ou G : ERREUR."
  Aff_TEXTE_EEPROM(106,39); Terminer(6); // " S ou s : Sauvegarder la configuration."
  Aff_TEXTE_EEPROM(145,10); Aff_TEXTE_EEPROM(155,10); Aff_TEXTE_EEPROM(128,17);
  // Ci-dessus : " R ou r : Restituer la configuration."
  Terminer(8); Aff_TEXTE_EEPROM(164,33);  // " M ou m : Morse sonore OUI / NON."
  Terminer(12); Aff_TEXTE_EEPROM(197,41); // " V ou v : Vitesse du débit du code Morse."
  Terminer(4); Aff_TEXTE_EEPROM(238,21); Terminer(24); // " T ou t : Mode TEXTE."
  Aff_TEXTE_EEPROM(259,7); Aff_TEXTE_EEPROM(245,8); Aff_TEXTE_EEPROM(266,8);
  // Ci-dessus :" L ou l : Mode LETTRES."
  Terminer(22); Aff_TEXTE_EEPROM(274,35); // " C ou c : Affiche la CONFIGURATION."
  Terminer(10); Aff_TEXTE_EEPROM(309,41); Terminer(4); // " P ou p : Espace disponible sous la PILE."
  Serial.print(" G ou g : "); Aff_TEXTE_EEPROM(384,23); Serial.print('.'); // "GROUPE d'identification."
  Terminer(11); Aff_TEXTE_EEPROM(350,34); Terminer(11); // " & ou 1 : Passer en Mode CRYPTAGE."
  Aff_46_etoiles(); Serial.println(); Serial.println();}

void POINT() {// Génère automatiquement le silence de séparation des éléments.
  if (Morse) {
      int T_element = 84; if (Rapide) T_element = 57;
      digitalWrite(BUZZER,HIGH); delay(T_element); digitalWrite(BUZZER,LOW); delay(T_element);}
  Serial.print('.');}
 
void TRAIT() {// Génère automatiquement le silence de séparation des éléments.
  if (Morse) {
      int T_element = 252; if (Rapide) T_element = 171; digitalWrite(BUZZER,HIGH); delay(T_element);
      digitalWrite(BUZZER,LOW); delay(T_element);}
  Serial.print('-');}

void TRANSCODE() {// Génère en Morse le Caractère représenté par l'ordre de la lettre dans l'alphabet.
 byte Lettre, Nb_Elements;
 int T_element = 252; if (Rapide) T_element = 171; Lettre = CODE_MORSE[Caractere]; N_ESPACES(2);
 Nb_Elements = Lettre >> 5; // Isole le nombre d'éléments par cinq décalages à droite.
 for (Nb_Elements; Nb_Elements > 0; Nb_Elements--) {
   if ((Lettre & 1) > 0) TRAIT(); else POINT(); Lettre = Lettre >> 1;}
 delay (T_element); Serial.println();} // Silence de 0,4S avec le délai entre éléments.

//====================================================================================================

void Faciliter_la_manipulation() {
  switch(Caractere) {
    case ',' : {Caractere = '?'; break;}
    case '1' : {if (!Mode_Initialisation) Caractere = '&'; break;}
    case 224 : {Caractere = 'a'; break;} case 226 : {Caractere = 'a'; break;}
    case 232 : {Caractere = 'e'; break;} case 233 : {Caractere = 'e'; break;}
    case 234 : {Caractere = 'e'; break;} case 235 : {Caractere = 'e'; break;}
    case 238 : {Caractere = 'i'; break;} case 239 : {Caractere = 'i'; break;}
    case 244 : {Caractere = 'o'; break;} case 231 : {Caractere = 'c'; break;}
    case 251 : {Caractere = 'u'; break;} case 252 : {Caractere = 'u'; break;}
    case '.' : Caractere = ' ';}} // Le '.' sera ignoré comme l'espace.

void Attendre_une_Reponse_sur_RX() {
  byte I = 0; LGR_chaine = 0; // Pointeur dans Chaine_Memorisee.
  while(!Serial.available()) { // Attendre un premier caractere.
     //================= Chronométrage pour le délai écoulé ================
     if (millis() - TOP_chronometre > 50)  {
        TOP_chronometre = millis(); digitalWrite(LED_bleue,!digitalRead(LED_bleue));}}
     //---------------------------------------------------------------------
  while(Serial.available()) {
     Caractere = Serial.read();
     Faciliter_la_manipulation();
     if ((Caractere > 96) && (Caractere < 123)) Caractere = Caractere & 223; // Bien moins couteux
              // en octets que Chaine_Memorisee.toUpperCase() pour forcer la chaine en MAJUSCULES.
     if (LGR_chaine < (Taille + 1)) {
        Chaine_Memorisee[LGR_chaine] = Caractere; LGR_chaine++; delay(5);}} // Délai impératif.
  Chaine_Memorisee[LGR_chaine] = '\0'; // Placer le délimiteur de fin de texte.
  //----------------------------------------------------------------------------------------
  if (Mode_COMMANDE) {AFF_crochet_ouvrant();
     while(Chaine_Memorisee[I] != '\0') {Serial.print(char(Chaine_Memorisee[I])); I++;}
     AFF_crochet_fermant(); Serial.println();}}

void Attendre_une_chaine_sur_RX() {Afficher_la_Commande(); Attendre_une_Reponse_sur_RX();}

//============================ Traite l'affichage des ERREURS  ============================

void Aff_texte_avant_erreur() {if (Texte) {
     Serial.println(); Serial.print('<');
     for(byte I = 0; I < Index; I++) Serial.print(Chaine_Memorisee[I]); Aff_TEXTE_EEPROM(584,2);}} // "> "

void Aff_ERREUR(byte Num_ERR) {
  switch (Num_ERR) {
     case 1 : { AFF_LF_Espace(); Aff_TEXTE_EEPROM(568,22); // " Attention : LGR > MAX."
                Index = 60; Aff_texte_avant_erreur(); Index = 0;
                L_autorisee = 255; // Pour éviter autant de BIPs que de dépassement.
                Compteur_pour_separer = 255; Compteur_pour_ligne = 255; break;}               
     case 2 : {Aff_texte_Caractere(); Aff_de_puis_Espace(); Aff_TEXTE_EEPROM(22,8);
               Aff_texte_invalide(); break;} // Ci-dessus : " Caractère de COMMANDE invalide."
     case 3 : {if (Texte) {
                   Aff_texte_avant_erreur(); Index = LGR_chaine;} // Ne pas envoyer la suite.    
                   Affiche_le_caractere_entre_crochets(); Aff_texte_Caractere();
                   Aff_de_puis_Espace(); Aff_TEXTE_EEPROM(375,8); Aff_texte_invalide(); break;}
                   // Ci-dessus : " Caractere de CRYPTAGE invalide."               
     case 4 : {Serial.println(); Aff_ROTOR(); Aff_texte_invalide(); break;}
     case 5 : {Serial.println(); Aff_Index(); Aff_texte_invalide(); break;}
     case 6 : {AFF_LF_Espace(); Aff_TEXTE_EEPROM(266,6); Aff_texte_invalide(); break;} // "LETTRE invalide."
     case 7 : {Serial.println(); Aff_TEXTE_EEPROM(516,7); Aff_TEXTE_EEPROM(193,3); Aff_TEXTE_EEPROM(522,10);
               break;} // " FICHES NON croisées.
     case 8 : {AFF_LF_Espace(); Aff_TEXTE_EEPROM(266,6); Aff_TEXTE_EEPROM(476,15); break;}
               // "LETTRE déjà utilisée"
     case 9 : {Serial.println(); Aff_TEXTE_EEPROM(590,20); break;} // " Un seul caractère !"
     case 10 : {AFF_LF_Espace(); Aff_TEXTE_EEPROM(661,7);// "Option ");
                Aff_texte_invalide(); break;}}
  Serial.println(); BIP();}

//======================== Traite l'affichage de la Configuration =========================

void Affiche_le_tableau_des_fiches() {
  boolean Passe_a_la_ligne = true; byte Num_Fiche = 1;
  for (byte I = 0; I < 20; I++) {
       Aff_Fiche_num(); if(Num_Fiche < 10) Serial.print('0'); Serial.print(Num_Fiche); Fleche();
       AFF_crochet_ouvrant(); Serial.print(FICHES_croisees[I]); Serial.print(" et ");
       Serial.print(FICHES_croisees[I+1]); AFF_crochet_fermant(); N_ESPACES(4); Num_Fiche++; I++;
       Passe_a_la_ligne = !Passe_a_la_ligne; if (Passe_a_la_ligne) Serial.println();}}

void Transmet_le_GROUPE_de_tete() {
  byte Compteur = 0;
  for (byte I = 0; I < 12; I++) {
       Serial.print(GROUPE_de_tete[I]); Compteur++; if (Compteur == 3) {N_ESPACES(1); Compteur = 0;}}}

void Aff_Num_Rotor(byte Num_Rotor) {
  byte Ordre;
  switch (Num_Rotor) {
     case 1 : { Ordre = Rotor_de_Droite; break;}
     case 2 : { Ordre = Rotor_du_Centre; break;}  
     case 3 : Ordre = Rotor_de_Gauche;}
  switch (Ordre) {
     case 1 : { Serial.print('I'); N_ESPACES(2); break;}
     case 2 : { Serial.print("II"); N_ESPACES(1); break;}  
     case 3 : { Serial.print("III"); break;}
     case 4 : { Serial.print("IV"); N_ESPACES(1); break;}
     case 5 :  Serial.print('V'); N_ESPACES(2);}}

void Attendre_un_chiffre() {
  Fleche(); Attendre_une_Reponse_sur_RX(); Caractere = (byte(Caractere)- 48);}

void Attendre_un_nombre() {
  Fleche(); Attendre_une_Reponse_sur_RX(); Nombre = atoi(Chaine_Memorisee);}
  
void Affiche_la_CONFIGURATION_machine() {
  Affiche_triple_fleche_ouvrantes(); Aff_TEXTE_EEPROM(295,13); Affiche_triple_fleche_fermantes();
  Serial.println(); Aff_Rotor_de_gauche(); Affiche_Egal();  Aff_Num_Rotor(3);
  Aff_Index(); Aff_de_Gauche(); Affiche_Egal();  if(Index_de_Gauche < 10) N_ESPACES(1);
  Serial.print(Index_de_Gauche); Aff_Orientation(); Affiche_Egal(); Serial.println(Orientation_de_Gauche); 
  Aff_Rotor_du_centre(); Affiche_Egal();  Aff_Num_Rotor(2);
  Aff_Index(); Aff_du_Centre(); Affiche_Egal(); if(Index_du_Centre < 10) N_ESPACES(1);
  Serial.print(Index_du_Centre); Aff_Orientation(); Affiche_Egal();
  Serial.println(Orientation_du_Centre);
  Aff_Rotor_de_droite(); Affiche_Egal(); Aff_Num_Rotor(1);
  Aff_Index(); Aff_de_Droite(); Affiche_Egal(); if(Index_de_Droite < 10) N_ESPACES(1);
  Serial.print(Index_de_Droite); Aff_Orientation(); Affiche_Egal(); Serial.println(Orientation_de_Droite);
  N_ESPACES(1); Aff_TEXTE_EEPROM(466,10); Affiche_Egal(); Serial.print(REFLECTEUR); Serial.println();
  Affiche_le_tableau_des_fiches(); Aff_Groupe_Identification(); // "GROUPE d'identification"
  Affiche_Egal(); Transmet_le_GROUPE_de_tete(); Serial.println(); Serial.println();}

//=========================== Traite l'INITIALISATION d'Énigma ============================

void Choisir_un_Rotor(byte Num_Rotor) {
  byte  Num_ERR = 4; // Pour entrer dans la boucle.
  while (Num_ERR > 0){
    switch(Num_Rotor) {
       case 1 : {Aff_Rotor_de_gauche(); break;}
       case 2 : {Aff_Rotor_du_centre(); break;}
       case 3 : Aff_Rotor_de_droite();}
    Num_ERR = 0; Attendre_un_chiffre(); if (LGR_chaine > 1) Aff_ERREUR(1);
    if ((Caractere < 1) || (Caractere > 5)) {Num_ERR = 4; Aff_ERREUR(4);}
    else switch(Num_Rotor) {
            case 1 : {Rotor_de_Gauche = Caractere; break;}
            case 2 : {Rotor_du_Centre = Caractere; break;}
            case 3 : Rotor_de_Droite = Caractere;}}}

void Choix_des_ROTORS() {
  Serial.println(); Choisir_un_Rotor(1); Choisir_un_Rotor(2); Choisir_un_Rotor(3);}

void Configurer_en_interne(byte Num_Rotor) {
  byte  Num_ERR = 5; // Pour entrer dans la boucle.
  while (Num_ERR > 0){
    Num_ERR = 0; Serial.print(" Indexation interne du");
    switch(Num_Rotor) {
       case 1 : {Aff_Rotor_de_gauche(); break;}
       case 2 : {Aff_Rotor_du_centre(); break;}
       case 3 : Aff_Rotor_de_droite();}
    Attendre_un_nombre();
    if ((Nombre < 1) || (Nombre > 26)) {Num_ERR = 5; Aff_ERREUR(5);}
    else switch(Num_Rotor) {
            case 1 : {Index_de_Gauche = Nombre; break;}
            case 2 : {Index_du_Centre = Nombre; break;}
            case 3 : Index_de_Droite = Nombre;}}} 

void Configuration_interne_des_ROTORS() {
  Configurer_en_interne(1); Configurer_en_interne(2); Configurer_en_interne(3);}

void Orienter(byte Num_Rotor) {
  byte  Num_ERR = 6; // Pour entrer dans la boucle.
  while (Num_ERR > 0){ Num_ERR = 0; Serial.print(" Orientation");
    switch(Num_Rotor) {
       case 1 : {Aff_Rotor_de_gauche(); break;}
       case 2 : {Aff_Rotor_du_centre(); break;}
       case 3 : Aff_Rotor_de_droite();}     
    Fleche(); Attendre_une_Reponse_sur_RX(); if (LGR_chaine > 1) Aff_ERREUR(1);
    Caractere = Chaine_Memorisee[0];
    if ((Caractere < 'A') || (Caractere > 'Z')) {Num_ERR = 6; Aff_ERREUR(6);}
    else switch(Num_Rotor) {
           case 1 : {Orientation_de_Gauche = Caractere; break;} 
           case 2 : {Orientation_du_Centre = Caractere; break;}
           case 3 : Orientation_de_Droite = Caractere;}}}

void Initialiser_la_position_des_Rotors() {Orienter(1); Orienter(2); Orienter(3);}
           
void Initialiser_le_Reflecteur() {
  byte  Num_ERR = 6; // Pour entrer dans la boucle.
  while (Num_ERR > 0){ Num_ERR = 0; Serial.println(); N_ESPACES(1); Aff_TEXTE_EEPROM(466,10); Fleche();
    Attendre_une_Reponse_sur_RX(); if (LGR_chaine > 1) Aff_ERREUR(1);
    Caractere = Chaine_Memorisee[0];
    if ((Caractere < 'B') || (Caractere > 'C')) {Num_ERR = 6; Aff_ERREUR(6);}
    REFLECTEUR = Caractere;}}

boolean Caractere_alphabetique() {return ((Caractere >= 'A') && (Caractere <= 'Z'));}

void Initialiser_le_GROUPE_de_tete() {
  byte Num_ERR = 1; // Pour entrer dans la boucle.
  while (Num_ERR > 0){Num_ERR = 0; Serial.println(); N_ESPACES(1); Aff_Groupe_Identification(); Fleche();
    Attendre_une_Reponse_sur_RX(); if (LGR_chaine > 12) {Num_ERR = 1; Aff_ERREUR(1);}
    if (Num_ERR == 0) {
       GROUPE_de_tete[0] = Orientation_de_Gauche;
       GROUPE_de_tete[1] = Orientation_du_Centre;
       GROUPE_de_tete[2] = Orientation_de_Droite;
       for (byte I = 0; I < 9; I++) {
           Caractere = Chaine_Memorisee[I]; Num_ERR = !Caractere_alphabetique();
           if (Num_ERR == 0) GROUPE_de_tete[I+3] = Caractere; else {Aff_ERREUR(6); I = 9;}}}}}

void Verifie_non_utilisee() {if (Lettre_utilisee[Caractere - 65]) {Num_ERR = 7; Aff_ERREUR(8);}}

void Verifie_le_caractere(byte Index) {
  Caractere = Chaine_Memorisee[Index]; Num_ERR = !Caractere_alphabetique();
  if (Num_ERR > 0) Aff_ERREUR(6); else Verifie_non_utilisee();}

void Verifie_la_fiche() {
  if (Chaine_Memorisee[0] == Chaine_Memorisee[1]) {Num_ERR = 7; Aff_ERREUR(7);} // " FICHES NON croisées."
  if (Num_ERR == 0) {if (LGR_chaine > 2) {Num_ERR = 1; Aff_ERREUR(1);}
                     else {if (LGR_chaine < 2) {Num_ERR = 9; Aff_ERREUR(9);}
                           else {Verifie_le_caractere(0); if (Num_ERR == 0) Verifie_le_caractere(1);}}}}

void Liste_deja_utilisees() {
  N_ESPACES(1); AFF_crochet_ouvrant();
  for (byte I = 0; I < 26; I++) if (Lettre_utilisee[I]) Serial.print(char(I+65)); Serial.println(']');}

void Initialiser_le_tableau_des_FICHES() {
  byte Num_Fiche = 1;
  for (byte I = 0; I < 26; I++) Lettre_utilisee[I] = false; // Aucune lettre n'est encore utilisée.
  for (byte I = 0; I < 20; I++) {// Traiter 10 fiches croisées.
      Num_ERR = 1; // Pour entrer dans la boucle.
      while (Num_ERR > 0){Num_ERR = 0; Aff_Fiche_num(); 
            Serial.print(Num_Fiche); Fleche(); Attendre_une_Reponse_sur_RX(); Verifie_la_fiche();
            if (Num_ERR == 0) {FICHES_croisees[I] = Chaine_Memorisee[0];
                Lettre_utilisee[Chaine_Memorisee[0] - 65] = true; FICHES_croisees[I+1] = Chaine_Memorisee[1];
                Lettre_utilisee[Chaine_Memorisee[1] - 65] = true; Liste_deja_utilisees(); Num_Fiche++; I++;}}}}

void Initialiser_Rotors_et_Reflecteur() {Choix_des_ROTORS(); Configuration_interne_des_ROTORS();
  Initialiser_la_position_des_Rotors(); Initialiser_le_Reflecteur();}
 
void Mode_INITIALISATION() {Serial.println();
  Mode_Initialisation = true;
  Affiche_triple_fleche_ouvrantes(); Aff_TEXTE_EEPROM(86,19); // " Mode INITIALISATION"
  Affiche_triple_fleche_fermantes(); Serial.println(); Confirmer();
  if (OUI) {
     Initialiser_Rotors_et_Reflecteur(); Initialiser_le_tableau_des_FICHES(); Initialiser_le_GROUPE_de_tete();
     Serial.println(); Mode_Initialisation = false;}} 

//============================ Traite l'affichage de la PILE  =============================

int SRAM_LIBRE() { // Fonction qui retourne la taille de SRAM disponible.
  extern int __heap_start, *__brkval; // Déclaration des deux pointeurs dédiés.
  byte BIDON; // Dernière variable allouée, donc occupe le "haut" du TAS.
  if (__brkval == 0) {return (int) &BIDON -(int) &__heap_start;}
  else {return (int) &BIDON -(int) __brkval;}}

void Affiche_espace_sous_la_PILE() {
  Serial.print(" RAM DYNAMIQUE (PILE - TAS)  =  "); Serial.print (SRAM_LIBRE());
  Serial.println(" octets.");}

//================================== Traiter un CRYPTAGE ==================================

void Crypte_le_caractere() {// Reçoit l'ordre alphabétique du caractère crypté.
  if (Caractere == 26) Caractere = 1; else Caractere++;} // Traverse le composant virtuel.
 
void Verification_du_cryptage() {
  ERREUR = 0; Caractere = Chaine_Memorisee[Index];
  if (LGR_chaine > L_autorisee) ERREUR = 1;
  if ((Caractere < 'A') || (Caractere > 'Z')) ERREUR = 3;
  if ((Caractere == ' ') && (!Texte)) ERREUR = 3;
  if (ERREUR > 0) Aff_ERREUR(ERREUR);}

void Traiter_le_Caractere() {
  if (Caractere == '&') Mode_COMMANDE = true;
  if (Mode_COMMANDE) Aff_retour_Commande();
  else {Verification_du_cryptage();
       if (ERREUR == 0) {
          Caractere = (byte(Chaine_Memorisee[Index])) - 64; Crypte_le_caractere();
          Affiche_caractere_Num();}}}

void Traiter_le_TEXTE() {// Index = 0 en entrée de cette procédure.
  for(Index; Index < LGR_chaine; Index++) {
      Caractere = Chaine_Memorisee[Index];
      if (!(Caractere == ' ')) {  // Ignorer ' ' en mode TEXTE.
         Traiter_le_Caractere(); Compteur_pour_separer++; Compteur_pour_ligne++;
      if (Compteur_pour_separer == 6) {N_ESPACES(1); Compteur_pour_separer = 0;}
      if (Compteur_pour_ligne == 54) {Serial.println(); Compteur_pour_ligne = 0;}}}}

void Traiter_le_CRYPTAGE() { Index = 0; // Impératif pour Verification_du_cryptage().
  if (Texte) {L_autorisee = Taille; Traiter_le_TEXTE();}
  else {Caractere = Chaine_Memorisee[0]; L_autorisee = 1;
        if (Caractere != '&') Verification_du_cryptage();
        if (ERREUR == 0) {
           if (Caractere == '&') {
               Mode_COMMANDE = true; Serial.println(); Aff_retour_Commande();}
           else {AFF_crochet_ouvrant(); Serial.print(char(Caractere)); 
                 Aff_TEXTE_EEPROM(619,11); // "] devient ["
                 Traiter_le_Caractere(); AFF_crochet_fermant(); TRANSCODE();}}
        else  ERREUR = 0;}} // Impératif pour reprendre la main avec '&'.

//============================ Traite la SAUVEGARDE en EEPROM  ============================

void Sauvegarde_Initialisation() {Serial.println(" Traite la SAUVEGARDE.");}

//====================== Traite le Rchargement d'une configuration  ======================

void Restitue_Initialisation() {Serial.println(" Traite le RECHARGEMENT.");}

//================================= Traiter une COMMANDE ==================================

void Analyseur_syntaxique() {
  ERREUR = 0; L_autorisee = 1; Caractere = Chaine_Memorisee[0]; if (Caractere == 'E') L_autorisee = 2;
  if (LGR_chaine > L_autorisee) ERREUR = 1;
  if (ERREUR == 0) {
     switch (Caractere) { // Refuse les caractères de commande invalides.
         case '?' : {break;} case 'I' : {break;} case 'S' : {break;} case 'R' : {break;} case 'M' : {break;}
         case 'V' : {break;} case 'T' : {break;} case 'L' : {break;} case 'C' : {break;} case 'P' : {break;}
         case 'G' : {break;} case 'E' : {break;} case '&' : {break;}
     default : ERREUR = 2;}}
     if (ERREUR > 0) Aff_ERREUR(ERREUR);}

void Affiche_le_mode(byte Mode) {
  Affiche_passage_au_mode(); // " Passage en Mode "
  switch (Mode) {
    case 1 : {if (Mode_COMMANDE) {Aff_TEXTE_EEPROM(22,9); Serial.println('.');} // "Commande."
              else {Aff_TEXTE_EEPROM(375,9); BIP(); Serial.println();} break;} // "CRYPTAGE."
    case 2 : {Aff_TEXTE_EEPROM(253,7); Serial.println("(60 caracteres MAXI.)"); break;}
    case 3 : {Aff_TEXTE_EEPROM(266,8); Serial.println();}}} // "LETTRES."

void Traiter_la_COMMANDE() {
  byte I = 0; // Pointer le début de la chaine TEXTE.
  Analyseur_syntaxique(); // Vérifie les COMMANDES.
  if (ERREUR == 0)
     switch (Caractere) {
       case '?' : {Affiche_le_MENU(); break;}
       case 'I' : {Mode_INITIALISATION(); break;}
       case 'S' : {Sauvegarde_Initialisation(); break;}
       case 'R' : {Restitue_Initialisation(); break;} // Ci-dessous " Morse sonore = "
       case 'M' : {Morse = !Morse; Serial.print("(BIP)"); N_ESPACES(1); Aff_TEXTE_EEPROM(173,14);
                   Serial.print("= "); Aff_OUI_NON(Morse); Serial.println(); break;}
       case 'V' : {Rapide = !Rapide;{ Aff_TEXTE_EEPROM(206,9); Aff_TEXTE_EEPROM(232,5);
                   Affiche_Egal();} // Ci-dessus " Vitesse Morse "
                   if (Rapide) {Aff_TEXTE_EEPROM(0,7); Serial.println();} // "Rapide."
                   else {Aff_TEXTE_EEPROM(7,6); Serial.println();} break;} // "Lente.")
       case 'T' : {Texte = true; Affiche_le_mode(2); break;}
       case 'L' : {Texte = false; Affiche_le_mode(3); break;}
       case 'C' : {Serial.println(); Affiche_la_CONFIGURATION_machine(); break;}
       case 'P' : {Affiche_espace_sous_la_PILE(); break;}
       case '&' : {Mode_COMMANDE = false; Compteur_pour_separer = 0; Compteur_pour_ligne = 0;
                   Serial.println(); Affiche_passage_au_mode(); Aff_CRYPTAGE();
                   Serial.println();}}}
   
  

