
/* Démonstrateur qui reprend intégralement le contenu du démonstrateur P07 et qui ajoute dans le
   cryptage la présence du Rotor de Droite et qui le fait tourner à chaque lettre frappée.
   
   Changement de stratégie pour le bruiteur : La commande 'M' n'affecte que le Morse sonore. Pour
   suspendre ou valider le BIP d'erreur la commande 'B' a été introduite.
   
   UTILISATION : ATTENTION : Il faut au préalable loger en EEPROM la nouvelle version de codage des rotors
                 en utilisant P00C_Initialiser_EEPROM.ino et écrire ces données en activant le Moniteur.

                    >>>>>> Manipulations proposées dans le didacticiel en page 37.
   
            ******************************************************************
            * Ce démonstrateur constitue le noyau du programme d'utilisation *
            * et sera complété au fur et à mesure du développement.          *
            ******************************************************************

Compilateur v1.7.9 :                                      PGM                   RAM dynamique
P08 avant toute modification :                        9168 Octets (29%)        544 Octets (26%)
Affichage fenêtre des rotors et divers correctifs :   9586 Octets (31%)        561 Octets (27%)
Criptage avec D et R et indexation :                  9740 Octets (31%)        613 Octets (29%) */

//--------------------- Définition des broches d'E/S ------------------------
#define BUZZER 4
#define LED_bleue 5

//------------------------ Constantes du programme --------------------------
#define Version "15/02/2024" // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Vitesse_USB 57600  // Taux de transfert sur la ligne USB @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Taille 60 // Longueur maximale de la chaine de dialogue @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

#define Direct true // Traverse le Rotor 1 en Direct @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Reflechi false // Traverse le Rotor 1 en Reflechi @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Indexer_en_Direct true // Fait tourner le Rotor 1 @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Indexer_en_Reflechi false // Fait tourner le Rotor 1 @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Rotation_en_Direct false // Fait tourner le Rotor 1 @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Rotation_en_Reflechi false // Fait tourner le Rotor 1 @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

// Taille réservée pour une chaine de 60 plus le délimiteur.
// Inutile de réserver plus de 60 caractère car la limite du FIFO USB est de 63 caractères.

//------------------------- Variables du programme --------------------------

     // Rapide : Cadence de la manipulation Morse.
     // Texte : Mode TEXTE au lieu du mode LETTRES.
boolean Bip, Groupe_identification, Mode_COMMANDE, Mode_Initialisation, Morse, OUI, Rapide, RESET, Texte;

     // Caractere contiendra l'ordre alphabétique du caractère en cours de traitement.
     // Index : Pointeur pour le cryptage dans la chaine TEXTE.
byte Caractere, Compteur_pour_separer, Compteur_pour_ligne, Index, Index_de_Droite,
     Index_de_Gauche, Index_du_Centre, LGR_chaine, L_autorisee, Num_ERR, Rotor_de_Droite, Rotor_de_Gauche,
     Rotor_du_Centre;

char Orientation_de_Droite, Orientation_de_Gauche, Orientation_du_Centre, Phase_de_Droite,
     Phase_de_Gauche, Phase_du_Centre, REFLECTEUR ;

int Nombre, PTR;

unsigned long TOP_chronometre;

byte CODE_MORSE[27] = {66,129,133,97,32,132,99,128,64,142,101,130,67,65,
                       103,134,139,98,96,33,100,129,102,137,141,131};
     // NOTE : Pour simplifier le codage, l'indice 0 ne sera jamais utilisé et on aura
     //        Indice 1 pour A, indide 2 pour B etc.

byte Rotor_de_DroiteD[26] = {};
byte Rotor_de_DroiteR[26] = {};

char GROUPE_de_tete[12] = {};
char FICHES_croisees[20] = {};
boolean Lettre_utilisee[26];

char Chaine_Memorisee[Taille+2]; // Tampon mémoire pour la chaine entrante. Taille + 2 car il faut un
                          // emplacement pour le délimiteur et un emplacement pour "pouvoir déborder".

//-------------------------------------------------------------------------------------

void setup() {
  Serial.begin(57600); Serial.println(); Aff_TEXTE_EEPROM(439,12); //"Version du "
  Serial.println(Version);
  pinMode(LED_bleue, OUTPUT); digitalWrite(LED_bleue,LOW); pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER,LOW);
  //--------------------------- Initialisation des variable ---------------------------
  Bip = true; L_autorisee = 1; Mode_COMMANDE = true; Mode_Initialisation = false, Morse = false;
  Rapide = false; Texte = false; TOP_chronometre = 0;
  Affiche_le_MENU(); RESET = true; Restitue_Initialisation(); RESET = false;}
 
void loop() {
  Attendre_une_chaine_sur_RX();
  if (Mode_COMMANDE) Traiter_la_COMMANDE(); else Traiter_le_CRYPTAGE();}

//==================== Procédures Écriture / Lecture EEPROM ====================
byte Lire_Octet_EEPROM(int Pointeur) {
  return eeprom_read_byte((unsigned char *) Pointeur);}

byte Ecrire_Octet_EEPROM(int Pointeur, byte OCTET) {
  eeprom_write_byte((unsigned char*) Pointeur,OCTET); PTR++;}

void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {
    for (byte I=0; I<Longueur; I++) {Serial.print(char(Lire_Octet_EEPROM(PTR))); PTR++;}}

//======================== Procédures de servitude =============================

void BIP() {if (Bip) {digitalWrite(BUZZER,HIGH); delay(150); digitalWrite(BUZZER,LOW);}}

void N_ESPACES(byte Nb) {for(byte N = 0; N < Nb; N++)  Serial.print(' ');}
void Aff_etoile() {Serial.print('*');}
void Separer() {Aff_46_etoiles(); Serial.println(); N_ESPACES(1); Aff_etoile();}

void Aff_46_etoiles() {for(byte N = 0; N < 46; N++) Aff_etoile();}
void Aff_BIP() {Serial.print(" BIP");}
void AFF_crochet_ouvrant() {Serial.print('[');}
void AFF_crochet_fermant() {Serial.print(']');}
void Aff_CRYPTAGE() {Aff_TEXTE_EEPROM(375,9);}
void Aff_de_Droite() {Aff_TEXTE_EEPROM(410,10);} 
void Aff_de_Gauche() {Aff_TEXTE_EEPROM(420,10);}
void Aff_de_puis_Espace() {Aff_TEXTE_EEPROM(420,3);}
void Aff_du_Centre() {Aff_TEXTE_EEPROM(430,10);}
void Aff_Fiche_num() {Aff_TEXTE_EEPROM(515,6); N_ESPACES(1); Aff_TEXTE_EEPROM(654,2);} // " FICHE nº"
void Aff_Index() {N_ESPACES(1); Aff_TEXTE_EEPROM(656,5); N_ESPACES(1);}
void AFF_LF_Espace() {Serial.println(); N_ESPACES(1);}
void Aff_O_N() {Serial.print("(O/N)");}
void Aff_Orientation() { N_ESPACES(1); Aff_TEXTE_EEPROM(451,11);}
void Aff_OUI_NON(boolean OUI) {if (OUI) Aff_TEXTE_EEPROM(187,3); else Aff_TEXTE_EEPROM(193,3);}
void Aff_Position_de_droite() {Aff_ROTOR(); Aff_de_Droite();}
void Aff_Position_de_gauche() {Aff_ROTOR(); Aff_de_Gauche();}
void Aff_Position_du_centre() {Aff_ROTOR(); Aff_du_Centre();}
void Aff_ROTOR() { N_ESPACES(1); Aff_TEXTE_EEPROM(462,5); N_ESPACES(1);}
void Aff_Rotor_de_droite() {Aff_ROTOR(); Aff_de_Droite();}
void Aff_Rotor_de_gauche() {Aff_ROTOR(); Aff_de_Gauche();}
void Aff_Rotor_du_centre() {Aff_ROTOR(); Aff_du_Centre();}
void Aff_texte_Caractere() {Aff_TEXTE_EEPROM(598,11);} // " Caractère "
void Aff_texte_invalide() {N_ESPACES(1); Aff_TEXTE_EEPROM(610,9);} // " invalide."
void Affiche_caractere_en_lettre() {Serial.print(char(Caractere + 64));}
void Affiche_Egal() {Serial.print(" = ");}
void Aff_Groupe_Identification() {N_ESPACES(1); Aff_TEXTE_EEPROM(384,23);} // "GROUPE d'identification"
void Affiche_passage_au_mode() {Aff_TEXTE_EEPROM(630,17);}
void Affiche_triple_fleche_ouvrantes() {Serial.print(" >>> ");}
void Affiche_triple_fleche_fermantes() {Serial.print(" <<<");}
void Fleche() {Aff_TEXTE_EEPROM(42,4);}

void Aff_fenetre_Rotors() {
  AFF_crochet_ouvrant(); Serial.print(Phase_de_Gauche); N_ESPACES(1); Serial.print(Phase_du_Centre);
  N_ESPACES(1); Serial.print(Phase_de_Droite); AFF_crochet_fermant();}

void Aff_retour_Commande() {Serial.println(); Serial.println(); Aff_TEXTE_EEPROM(646,8);
  Aff_TEXTE_EEPROM(639,8); Aff_TEXTE_EEPROM(22,10); Serial.println();} // " Retour au mode COMMANDES."

void Afficher_la_Commande() {
  if (Mode_COMMANDE) {Aff_TEXTE_EEPROM(21,9); Fleche();} // " COMMANDE -> "
  else {if (!Texte) {N_ESPACES(1); Aff_TEXTE_EEPROM(32,14);}}} // " Une lettre -> ")

void Affiche_le_caractere_entre_crochets() {
  AFF_crochet_ouvrant(); Serial.print(char(Caractere)); AFF_crochet_fermant();}

void Terminer(byte Nb) {
  for(byte N = 0; N < Nb; N++) N_ESPACES(1); Serial.println('*'); N_ESPACES(1); Serial.print('*');}

void Confirmer() {
  Serial.print(" Confirmer "); Aff_O_N(); Fleche(); Attendre_une_Reponse_sur_RX(); OUI = Caractere == 'O';}

void Affiche_le_MENU() {
   Serial.println(); N_ESPACES(1); Aff_etoile(); Separer();
  N_ESPACES(13); Aff_TEXTE_EEPROM(13,19); // "Menu des COMMANDES."
  Terminer(13); Separer(); Aff_TEXTE_EEPROM(45,31); // " ? ou , : Rappel des commandes."
  Terminer(14); Serial.print(" B ou b : ");
  Aff_BIP(); N_ESPACES(1); Aff_O_N(); Serial.print('.'); // "BIP (O/N)"
  Terminer(24); Aff_TEXTE_EEPROM(76,30); Terminer(15); // " I ou i : Mode INITIALISATION."
  Aff_TEXTE_EEPROM(532,36);  Terminer(9); // " E ou e suivi de R, F ou G : ERREUR."
  Aff_TEXTE_EEPROM(106,39); Terminer(6); // " S ou s : Sauvegarder la configuration."
  Aff_TEXTE_EEPROM(145,10); Aff_TEXTE_EEPROM(155,10); Aff_TEXTE_EEPROM(128,17);
  // Ci-dessus : " R ou r : Restituer la configuration."
  Terminer(8); Aff_TEXTE_EEPROM(164,33);  // " M ou m : Morse sonore OUI / NON."
  Terminer(12); Aff_TEXTE_EEPROM(197,41); // " V ou v : Vitesse du débit du code Morse."
  Terminer(4); Aff_TEXTE_EEPROM(238,21); Terminer(24); // " T ou t : Mode TEXTE."
  Aff_TEXTE_EEPROM(259,7); Aff_TEXTE_EEPROM(245,8); Aff_TEXTE_EEPROM(266,8);
  // Ci-dessus :" L ou l : Mode LETTRES."
  Terminer(22); Aff_TEXTE_EEPROM(274,35); // " C ou c : Affiche la CONFIGURATION."
  Terminer(10); Aff_TEXTE_EEPROM(309,41); Terminer(4); // " P ou p : Espace disponible sous la PILE."
  Aff_TEXTE_EEPROM(350,34); Terminer(11); // " & ou 1 : Passer en Mode CRYPTAGE."
  Aff_46_etoiles(); Serial.println();}

void POINT() {// Génère automatiquement le silence de séparation des éléments.
  if (Morse) {
      int T_element = 84; if (Rapide) T_element = 57;
      digitalWrite(BUZZER,HIGH); delay(T_element); digitalWrite(BUZZER,LOW); delay(T_element);}
  Serial.print('.');}
 
void TRAIT() {// Génère automatiquement le silence de séparation des éléments.
  if (Morse) {
      int T_element = 252; if (Rapide) T_element = 171; digitalWrite(BUZZER,HIGH); delay(T_element);
      digitalWrite(BUZZER,LOW); delay(T_element);}
  Serial.print('-');}

void TRANSCODE() {// Génère en Morse le Caractère représenté par l'ordre de la lettre dans l'alphabet.
 byte Lettre, Nb_Elements;
 int T_element = 252; if (Rapide) T_element = 171; Lettre = CODE_MORSE[Caractere - 65]; N_ESPACES(2);
 Nb_Elements = Lettre >> 5; // Isole le nombre d'éléments par cinq décalages à droite.
 for (Nb_Elements; Nb_Elements > 0; Nb_Elements--) {
   if ((Lettre & 1) > 0) TRAIT(); else POINT(); Lettre = Lettre >> 1;}
 delay (T_element); Serial.println();} // Silence de 0,4S avec le délai entre éléments.

//====================================================================================================

void Faciliter_la_manipulation() {
  switch(Caractere) {
    case ',' : {Caractere = '?'; break;}
    case '1' : {if (!Mode_Initialisation) Caractere = '&'; break;}
    case 224 : {Caractere = 'a'; break;} case 226 : {Caractere = 'a'; break;}
    case 232 : {Caractere = 'e'; break;} case 233 : {Caractere = 'e'; break;}
    case 234 : {Caractere = 'e'; break;} case 235 : {Caractere = 'e'; break;}
    case 238 : {Caractere = 'i'; break;} case 239 : {Caractere = 'i'; break;}
    case 244 : {Caractere = 'o'; break;} case 231 : {Caractere = 'c'; break;}
    case 251 : {Caractere = 'u'; break;} case 252 : {Caractere = 'u'; break;}
    case '.' : Caractere = ' ';}} // Le '.' sera ignoré comme l'espace.

void Attendre_une_Reponse_sur_RX() {
  byte I = 0; LGR_chaine = 0; // Pointeur dans Chaine_Memorisee.
  while(!Serial.available()) { // Attendre un premier caractere.
     //================= Chronométrage pour le délai écoulé ================
     if (millis() - TOP_chronometre > 50)  {
        TOP_chronometre = millis(); digitalWrite(LED_bleue,!digitalRead(LED_bleue));}}
     //---------------------------------------------------------------------
  while(Serial.available()) {
     Caractere = Serial.read();
     Faciliter_la_manipulation();
     if ((Caractere > 96) && (Caractere < 123)) Caractere = Caractere & 223; // Bien moins couteux
              // en octets que Chaine_Memorisee.toUpperCase() pour forcer la chaine en MAJUSCULES.
     if (LGR_chaine < (Taille + 1)) {
        Chaine_Memorisee[LGR_chaine] = Caractere; LGR_chaine++; delay(5);}} // Délai impératif.
  Chaine_Memorisee[LGR_chaine] = '\0'; // Placer le délimiteur de fin de texte.
  //----------------------------------------------------------------------------------------
  if (Mode_COMMANDE) {AFF_crochet_ouvrant();
     while(Chaine_Memorisee[I] != '\0') {Serial.print(char(Chaine_Memorisee[I])); I++;}
     AFF_crochet_fermant(); Serial.println();}}

void Attendre_une_chaine_sur_RX() {Afficher_la_Commande(); Attendre_une_Reponse_sur_RX();}

//============================ Traite l'affichage des ERREURS  ============================

void Aff_ERREUR(byte Num_ERR) {
  switch (Num_ERR) {
     case 1 : {AFF_LF_Espace(); Aff_TEXTE_EEPROM(568,22); // " Attention : LGR > MAX."
              if (!Mode_COMMANDE && Texte) {// Affiche le TEXTE saisi pour le cryptage.
                  Serial.println(); Serial.print('<');
                  for(byte I = 0; I < 60; I++) Serial.print(Chaine_Memorisee[I]); Serial.print('>');
                  L_autorisee = 255; // Pour éviter autant de BIPs que de dépassements.
                  Compteur_pour_separer = 0; Compteur_pour_ligne = 0;} break;}              
     case 2 : {Aff_texte_Caractere(); Aff_de_puis_Espace(); Aff_TEXTE_EEPROM(22,8);
               Aff_texte_invalide(); break;} // Ci-dessus : " Caractère de COMMANDE invalide."
     case 3 : {AFF_LF_Espace(); Aff_TEXTE_EEPROM(661,6); Aff_texte_invalide(); break;}
     case 4 : {Serial.println(); Aff_ROTOR(); Aff_texte_invalide(); break;}  // Ci-dessus "Option invalide."
     case 5 : {Serial.println(); Aff_Index(); Aff_texte_invalide(); break;}
     case 6 : {Serial.println(); Serial.print(Chaine_Memorisee[Index]); Aff_TEXTE_EEPROM(83,3);
               Aff_TEXTE_EEPROM(266,6); Aff_texte_invalide(); N_ESPACES(1); // "LETTRE invalide. "
               if (!Mode_COMMANDE && Texte) {// Affiche le texte ayant provoqué l'erreur.
                  Serial.print('['); for(byte I = 0; I < LGR_chaine; I++) {Serial.print(Chaine_Memorisee[I]);
                  if (I == (Index-1))Serial.print("] <");} Serial.println('>');
                  Index = LGR_chaine; // Pour éviter autant de BIPs que de dépassements.
                  Compteur_pour_separer = 0; Compteur_pour_ligne = 0;} break;} // "LETTRE invalide."
     case 7 : {Serial.println(); Aff_TEXTE_EEPROM(516,7); Aff_TEXTE_EEPROM(193,3); Aff_TEXTE_EEPROM(522,10);
               break;} // " FICHES NON croisées.
     case 8 : {AFF_LF_Espace(); Aff_TEXTE_EEPROM(266,6); Aff_TEXTE_EEPROM(476,15); break;}
               // "LETTRE déjà utilisée"
     case 9 : {Serial.println(); Aff_TEXTE_EEPROM(590,20);}} // " Un seul caractère !"
  Serial.println(); BIP();}

//======================== Traite l'affichage de la Configuration =========================

void Transmet_le_GROUPE_de_tete() {
  byte Compteur = 0;
  for (byte I = 0; I < 12; I++) {
       Serial.print(GROUPE_de_tete[I]); Compteur++; if (Compteur == 3) {N_ESPACES(1); Compteur = 0;}}}

void Affiche_le_tableau_des_fiches() {
  boolean Passe_a_la_ligne = true; byte Num_Fiche = 1;
  for (byte I = 0; I < 20; I++) {
       Aff_Fiche_num(); if(Num_Fiche < 10) Serial.print('0'); Serial.print(Num_Fiche); Fleche();
       AFF_crochet_ouvrant(); Serial.print(FICHES_croisees[I]); Serial.print(" et ");
       Serial.print(FICHES_croisees[I+1]); AFF_crochet_fermant(); N_ESPACES(4); Num_Fiche++; I++;
       Passe_a_la_ligne = !Passe_a_la_ligne; if (Passe_a_la_ligne) Serial.println();}}

void Aff_Num_Rotor(byte Num_Rotor) {
  byte Ordre;
  switch (Num_Rotor) {
     case 1 : { Ordre = Rotor_de_Droite; break;}
     case 2 : { Ordre = Rotor_du_Centre; break;}  
     case 3 : Ordre = Rotor_de_Gauche;}
  switch (Ordre) {
     case 1 : { Serial.print('I'); N_ESPACES(2); break;}
     case 2 : { Serial.print("II"); N_ESPACES(1); break;}  
     case 3 : { Serial.print("III"); break;}
     case 4 : { Serial.print("IV"); N_ESPACES(1); break;}
     case 5 :  Serial.print('V'); N_ESPACES(2);}}

void Attendre_un_chiffre() {
  Fleche(); Attendre_une_Reponse_sur_RX(); Caractere = (byte(Caractere)- 48);}

void Attendre_un_nombre() {
  Fleche(); Attendre_une_Reponse_sur_RX(); Nombre = atoi(Chaine_Memorisee);}
  
void Affiche_la_CONFIGURATION_machine() {
  Affiche_triple_fleche_ouvrantes(); Aff_TEXTE_EEPROM(295,13); Affiche_triple_fleche_fermantes();
  Serial.println(); Aff_Rotor_de_gauche(); Affiche_Egal();  Aff_Num_Rotor(3);
  Aff_Index(); Aff_de_Gauche(); Affiche_Egal();  if(Index_de_Gauche < 10) N_ESPACES(1);
  Serial.print(Index_de_Gauche); Aff_Orientation(); Affiche_Egal(); Serial.println(Orientation_de_Gauche); 
  Aff_Rotor_du_centre(); Affiche_Egal();  Aff_Num_Rotor(2);
  Aff_Index(); Aff_du_Centre(); Affiche_Egal(); if(Index_du_Centre < 10) N_ESPACES(1);
  Serial.print(Index_du_Centre); Aff_Orientation(); Affiche_Egal();
  Serial.println(Orientation_du_Centre);
  Aff_Rotor_de_droite(); Affiche_Egal(); Aff_Num_Rotor(1);
  Aff_Index(); Aff_de_Droite(); Affiche_Egal(); if(Index_de_Droite < 10) N_ESPACES(1);
  Serial.print(Index_de_Droite); Aff_Orientation(); Affiche_Egal(); Serial.println(Orientation_de_Droite);
  N_ESPACES(1); Aff_TEXTE_EEPROM(466,10); Affiche_Egal(); Serial.print(REFLECTEUR); Serial.println();
  Affiche_le_tableau_des_fiches(); Aff_Groupe_Identification(); // "GROUPE d'identification"
  Affiche_Egal(); Transmet_le_GROUPE_de_tete(); Serial.println(); Serial.println();}

//=========================== Traite l'INITIALISATION d'Énigma ============================

void Choisir_un_Rotor(byte Num_Rotor) {
  byte  Num_ERR = 4; // Pour entrer dans la boucle.
  while (Num_ERR > 0){
    switch(Num_Rotor) {
       case 1 : {Aff_Rotor_de_gauche(); break;}
       case 2 : {Aff_Rotor_du_centre(); break;}
       case 3 : Aff_Rotor_de_droite();}
    Num_ERR = 0; Attendre_un_chiffre(); if (LGR_chaine > 1) Aff_ERREUR(1);
    if ((Caractere < 1) || (Caractere > 5)) {Num_ERR = 4; Aff_ERREUR(4);}
    else switch(Num_Rotor) {
            case 1 : {Rotor_de_Gauche = Caractere; break;}
            case 2 : {Rotor_du_Centre = Caractere; break;}
            case 3 : Rotor_de_Droite = Caractere;}}}

void Choix_des_ROTORS() {
  Serial.println(); Choisir_un_Rotor(1); Choisir_un_Rotor(2); Choisir_un_Rotor(3);}

void Configurer_en_interne(byte Num_Rotor) {
  byte  Num_ERR = 5; // Pour entrer dans la boucle.
  while (Num_ERR > 0){
    Num_ERR = 0; Serial.print(" Indexation interne du");
    switch(Num_Rotor) {
       case 1 : {Aff_Rotor_de_gauche(); break;}
       case 2 : {Aff_Rotor_du_centre(); break;}
       case 3 : Aff_Rotor_de_droite();}
    Attendre_un_nombre();
    if ((Nombre < 1) || (Nombre > 26)) {Num_ERR = 5; Aff_ERREUR(5);}
    else switch(Num_Rotor) {
            case 1 : {Index_de_Gauche = Nombre; break;}
            case 2 : {Index_du_Centre = Nombre; break;}
            case 3 : Index_de_Droite = Nombre;}}} 

void Configuration_interne_des_ROTORS() {
  Configurer_en_interne(1); Configurer_en_interne(2); Configurer_en_interne(3);}

void Orienter(byte Num_Rotor) {
  byte  Num_ERR = 6; // Pour entrer dans la boucle.
  while (Num_ERR > 0){ Num_ERR = 0; Serial.print(" Orientation");
    switch(Num_Rotor) {
       case 1 : {Aff_Rotor_de_gauche(); break;}
       case 2 : {Aff_Rotor_du_centre(); break;}
       case 3 : Aff_Rotor_de_droite();}     
    Fleche(); Attendre_une_Reponse_sur_RX(); if (LGR_chaine > 1) Aff_ERREUR(1);
    Caractere = Chaine_Memorisee[0];
    if ((Caractere < 'A') || (Caractere > 'Z')) {Num_ERR = 6; Aff_ERREUR(6);}
    else switch(Num_Rotor) {
           case 1 : {Orientation_de_Gauche = Caractere; Phase_de_Gauche = Caractere; break;} 
           case 2 : {Orientation_du_Centre = Caractere; Phase_du_Centre = Caractere; break;}
           case 3 : {Orientation_de_Droite = Caractere; Phase_de_Droite = Caractere; }}}}

void Initialiser_la_position_des_Rotors() {Orienter(1); Orienter(2); Orienter(3);}
           
void Initialiser_le_Reflecteur() {
  byte  Num_ERR = 6; // Pour entrer dans la boucle.
  while (Num_ERR > 0){ Num_ERR = 0; Serial.println(); N_ESPACES(1); Aff_TEXTE_EEPROM(466,10); Fleche();
    Attendre_une_Reponse_sur_RX(); if (LGR_chaine > 1) Aff_ERREUR(1);
    Caractere = Chaine_Memorisee[0];
    if ((Caractere < 'B') || (Caractere > 'C')) {Num_ERR = 6; Aff_ERREUR(6);}
    REFLECTEUR = Caractere;}}

boolean Caractere_alphabetique() {return ((Caractere >= 'A') && (Caractere <= 'Z'));}

void Copie_les_Orientations_en_tete_du_GROUPE() { GROUPE_de_tete[0] = Orientation_de_Gauche;
  GROUPE_de_tete[1] = Orientation_du_Centre; GROUPE_de_tete[2] = Orientation_de_Droite;}

void Initialiser_le_GROUPE_de_tete() {
  byte Num_ERR = 1; // Pour entrer dans la boucle.
  while (Num_ERR > 0){Num_ERR = 0; Serial.println(); N_ESPACES(1); Aff_Groupe_Identification(); Fleche();
    Attendre_une_Reponse_sur_RX(); if (LGR_chaine > 9) {Num_ERR = 1; Aff_ERREUR(1);}
    if (Num_ERR == 0) {
       Copie_les_Orientations_en_tete_du_GROUPE();
       for (byte I = 0; I < 9; I++) {
           Caractere = Chaine_Memorisee[I]; Num_ERR = !Caractere_alphabetique();
           if (Num_ERR == 0) GROUPE_de_tete[I+3] = Caractere; else {Aff_ERREUR(6); I = 9;}}}}}

void Verifie_non_utilisee() {if (Lettre_utilisee[Caractere - 65]) {Num_ERR = 7; Aff_ERREUR(8);}}

void Verifie_le_caractere(byte Index) {
  Caractere = Chaine_Memorisee[Index]; Num_ERR = !Caractere_alphabetique();
  if (Num_ERR > 0) Aff_ERREUR(6); else Verifie_non_utilisee();}

void Verifie_la_fiche() {
  if (Chaine_Memorisee[0] == Chaine_Memorisee[1]) {Num_ERR = 7; Aff_ERREUR(7);} // " FICHES NON croisées."
  if (Num_ERR == 0) {if (LGR_chaine > 2) {Num_ERR = 1; Aff_ERREUR(1);}
                     else {if (LGR_chaine < 2) {Num_ERR = 9; Aff_ERREUR(9);}
                           else {Verifie_le_caractere(0); if (Num_ERR == 0) Verifie_le_caractere(1);}}}}

void Liste_deja_utilisees() {
  N_ESPACES(1); AFF_crochet_ouvrant();
  for (byte I = 0; I < 26; I++) if (Lettre_utilisee[I]) Serial.print(char(I+65)); Serial.println(']');}

void Initialiser_le_tableau_des_FICHES() {// On impose d'implanter dix fiches sur la machine.
  byte Num_Fiche = 1;
  for (byte I = 0; I < 26; I++) Lettre_utilisee[I] = false; // Aucune lettre n'est encore utilisée.
  for (byte I = 0; I < 20; I++) {// Traiter 10 fiches croisées.
      Num_ERR = 1; // Pour entrer dans la boucle.
      while (Num_ERR > 0){Num_ERR = 0; Aff_Fiche_num(); 
            Serial.print(Num_Fiche); Fleche(); Attendre_une_Reponse_sur_RX(); Verifie_la_fiche();
            if (Num_ERR == 0) {FICHES_croisees[I] = Chaine_Memorisee[0];
               Lettre_utilisee[Chaine_Memorisee[0] - 65] = true; FICHES_croisees[I+1] = Chaine_Memorisee[1];
               Lettre_utilisee[Chaine_Memorisee[1] - 65] = true; Liste_deja_utilisees();
               Num_Fiche++; I++;}}}}

void Initialiser_Rotors_et_Reflecteur() {Choix_des_ROTORS(); Configuration_interne_des_ROTORS();
  Initialiser_la_position_des_Rotors(); Initialiser_le_Reflecteur();}
 
void Mode_INITIALISATION() {Serial.println();
  Mode_Initialisation = true;
  Affiche_triple_fleche_ouvrantes(); Aff_TEXTE_EEPROM(86,19); // " Mode INITIALISATION"
  Affiche_triple_fleche_fermantes(); Serial.println(); Confirmer();
  if (OUI) {
     Initialiser_Rotors_et_Reflecteur(); Initialiser_le_tableau_des_FICHES();
     Initialiser_le_GROUPE_de_tete(); Serial.println(); Mode_Initialisation = false;}} 

//============================ Traite l'affichage de la PILE  =============================

int SRAM_LIBRE() { // Fonction qui retourne la taille de SRAM disponible.
  extern int __heap_start, *__brkval; // Déclaration des deux pointeurs dédiés.
  byte BIDON; // Dernière variable allouée, donc occupe le "haut" du TAS.
  if (__brkval == 0) {return (int) &BIDON -(int) &__heap_start;}
  else {return (int) &BIDON -(int) __brkval;}}

void Affiche_espace_sous_la_PILE() {
  Serial.print(" RAM DYNAMIQUE (PILE - TAS)  =  "); Serial.print (SRAM_LIBRE());
  Serial.println(" octets.");}

//================================== Traiter un CRYPTAGE ==================================

void Traiter_la_rotation_des_Rotors() {
  byte SAV_Octet;
  //-------------- Le Rotor de Droite tourne avant chaque codage --------------
  if (Phase_de_Droite == char(90)) Phase_de_Droite = 'A'; else Phase_de_Droite++;
  //----------------- Rotation systématique du Rotor de Droite ----------------
  if (Rotation_en_Direct) {SAV_Octet = Rotor_de_DroiteD[0]; // Shift à gauche.
      for (byte I=0; I<25; I++) Rotor_de_DroiteD[I] = Rotor_de_DroiteD[I+1];
      Rotor_de_DroiteD[25] = SAV_Octet;}
  if (Rotation_en_Reflechi) {SAV_Octet = Rotor_de_DroiteR[0]; // Shift à gauche.
      for (byte I=0; I<25; I++) Rotor_de_DroiteR[I] = Rotor_de_DroiteR[I+1];
      Rotor_de_DroiteR[25] = SAV_Octet;}
  //---------------------------------------------------------------------------
  } 

void Installe_les_rotors_et_le_Reflecteur() { // À chaque début de cryptage on réinstalle le total.
  byte SAV_Octet;  Clone_Phases_avec_Orientations(); // orientation fenêtre initialisée;
  // ATTENTION : Ici les rotors ne sont pas encore orientés ils sont en A.
  //------------- Mise en place sur la machine du rotor de droite -------------
  // ATTENTION ici on prend que le I qui est codé en EEPROM. @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  for (byte I = 0; I < 26; I++) Rotor_de_DroiteD[I] = Lire_Octet_EEPROM(I + 712);
  for (byte I = 0; I < 26; I++) Rotor_de_DroiteR[I] = Lire_Octet_EEPROM(I + 738);
  //------------------- Indexe en D et R le Rotor de Droite -------------------
  if (Indexer_en_Direct) {
     for (byte Nb = 0; Nb < (Index_de_Droite - 1); Nb++) {// Shift Nb fois à droite.
         SAV_Octet = Rotor_de_DroiteD[25];
         for (byte I=25; I>0; I--) Rotor_de_DroiteD[I] = Rotor_de_DroiteD[I-1];
         Rotor_de_DroiteD[0] = SAV_Octet;}}
  if (Indexer_en_Reflechi) {
      for (byte Nb = 0; Nb < (Index_de_Droite - 1); Nb++) {// Shift Nb fois à droite.
          SAV_Octet = Rotor_de_DroiteR[25];
          for (byte I=25; I>0; I--) Rotor_de_DroiteR[I] = Rotor_de_DroiteR[I-1];
          Rotor_de_DroiteR[0] = SAV_Octet;}}
   //--------------- Oriente le rotor de Droite dans la fenêtre ----------------
   }
   //---------------------------------------------------------------------------

void Crypte_le_caractere() {// Reçoit l'ordre alphabétique du caractère crypté.
  // >>>> REMARQUE : Durant tout le cryptage le caractère sera traité en "ordre alphabétique".
  // >>>> Ce n'est qu'avant sa "sortie vers l'ampoule virtuelle" qu'il sera traduit en "char".
  Traiter_la_rotation_des_Rotors();
  //------------------ Traverse en Direct le Rotor de Droite ------------------
  if (Direct) {Caractere = Caractere + Rotor_de_DroiteD[Caractere-1];
     if (Caractere > 26) Caractere = Caractere - 26;}
  //----------------- Traverse en Refléchi le Rotor de Droite -----------------
  if (Reflechi) {Caractere = Caractere + Rotor_de_DroiteR[Caractere-1];
     if (Caractere > 26) Caractere = Caractere - 26;}
  //---------------------------------------------------------------------------
  Caractere = Caractere + 64;} // Codage final du caractère en type [char] pour les affichages.
 
void Verification_du_cryptage() {
  Num_ERR = 0; Caractere = Chaine_Memorisee[Index];
  if (LGR_chaine > L_autorisee) Num_ERR = 1;
  if ((Caractere < 'A') || (Caractere > 'Z')) Num_ERR = 6;
  if ((Caractere == ' ') && (!Texte)) Num_ERR = 3;
  if (Num_ERR > 0) Aff_ERREUR(Num_ERR);}

void Traiter_le_Caractere() {
  if (Caractere == '&') {
      Mode_COMMANDE = true; // Ci-dessous : Affiche la fenêtre des rotors en sortie de saisie du TEXTE.
      if (Texte) {Serial.println(); Affiche_triple_fleche_ouvrantes(); Aff_fenetre_Rotors();}
      Aff_retour_Commande(); while (Serial.available()) Caractere = Serial.read(); Index = LGR_chaine;}
      // Ligne ci-dessus : Vide le tampon USB pour éviter une cascade de "comandes fantômes" si dans le
      //                   TEXTE on insère un &. Exemple "AAA&BBBB".
  else {Verification_du_cryptage(); // Traité si LGR <= LGR maximale. (Num_ERR = 1 si LGR > LGR maximale.)
       if (Num_ERR < 2) {Caractere = (byte(Chaine_Memorisee[Index])) - 64; Crypte_le_caractere();}}}

void Traiter_le_TEXTE() {// Index = 0 en entrée de cette procédure.
  for(Index; Index < LGR_chaine; Index++) {
      Caractere = Chaine_Memorisee[Index];
      if (Caractere != ' ') {// Ignorer ' ' en mode TEXTE.        (Num_ERR = 1 si tpGR > LGR maximale.)   
         Traiter_le_Caractere(); if ((!Mode_COMMANDE) && (Num_ERR < 2)) Serial.print(char(Caractere));
         Compteur_pour_separer++; Compteur_pour_ligne++;
         if (Compteur_pour_separer == 6) {N_ESPACES(1); Compteur_pour_separer = 0;}
         if (Compteur_pour_ligne == 54) {Serial.println(); Compteur_pour_ligne = 0;}}}}

void Traiter_le_CRYPTAGE() { Index = 0; // Impératif pour Verification_du_cryptage().
  if (Texte) {L_autorisee = Taille;
              if (Groupe_identification) {// Ne transmettre qu'une seule fois le GROUPE d'identification.
                 Transmet_le_GROUPE_de_tete();  Serial.println(); Groupe_identification = false;}
              Traiter_le_TEXTE();}
  else {Caractere = Chaine_Memorisee[0]; L_autorisee = 1;
        if (Caractere != '&') Verification_du_cryptage();
        if (Num_ERR == 0) {
           if (Caractere == '&') { Mode_COMMANDE = true; Aff_retour_Commande();}
           else {Traiter_le_Caractere(); Aff_fenetre_Rotors(); Affiche_triple_fleche_ouvrantes();
                AFF_crochet_ouvrant(); Serial.print(Chaine_Memorisee[0]);
                Aff_TEXTE_EEPROM(619,11); // "] devient ["
                Serial.print(char(Caractere)); AFF_crochet_fermant(); TRANSCODE();}}
        else  Num_ERR = 0;}} // Impératif pour reprendre la main avec '&'.

//=================== Traite l'éditeur d'ERREUR de configuration d'ÉNIGMA  ================

void Saisie_configuration_des_ROTORS() {Mode_Initialisation = true; 
  Initialiser_Rotors_et_Reflecteur(); Copie_les_Orientations_en_tete_du_GROUPE();
  Serial.println(); Affiche_la_CONFIGURATION_machine(); Mode_Initialisation = false;}
  
void Saisie_configuration_des_FICHES() {
   Serial.println(); Initialiser_le_tableau_des_FICHES(); Serial.println();
   Affiche_la_CONFIGURATION_machine();}

void Saisie_du_GROUPE_identification() {
  Initialiser_le_GROUPE_de_tete();Serial.println(); Affiche_la_CONFIGURATION_machine();}

void Verifie_la_reponse() {
  if (LGR_chaine < 2) {Num_ERR = 9; Aff_ERREUR(9);} // Affiche "Un seul caractère !"
  if (Num_ERR == 0) {
     Caractere = Chaine_Memorisee[1];
     switch (Caractere) {
         case 'R' : {Saisie_configuration_des_ROTORS(); break;}
         case 'F' : {Saisie_configuration_des_FICHES(); break;}
         case 'G' : {Saisie_du_GROUPE_identification(); break;}
     default : {Num_ERR = 3; Aff_ERREUR(3);}}}}

void Corriger_une_ERREUR() {
  Serial.println(); Affiche_triple_fleche_ouvrantes(); Serial.print("Corriger"); Aff_TEXTE_EEPROM(291,17); 
  N_ESPACES(1); Affiche_triple_fleche_fermantes(); Num_ERR = 0; Verifie_la_reponse();}

//============================ Traite la SAUVEGARDE en EEPROM  ============================
void Sauvegarde_Initialisation() {
  Serial.println(); Affiche_triple_fleche_ouvrantes(); Aff_TEXTE_EEPROM(116,28);
  Affiche_triple_fleche_fermantes(); Serial.println(); Confirmer();
  if (OUI) {PTR = 670; Ecrire_Octet_EEPROM(PTR, REFLECTEUR);
     Ecrire_Octet_EEPROM(PTR, Rotor_de_Gauche); Ecrire_Octet_EEPROM(PTR, Rotor_du_Centre);
     Ecrire_Octet_EEPROM(PTR, Rotor_de_Droite); Ecrire_Octet_EEPROM(PTR, Index_de_Gauche);
     Ecrire_Octet_EEPROM(PTR, Index_du_Centre); Ecrire_Octet_EEPROM(PTR, Index_de_Droite);
     Ecrire_Octet_EEPROM(PTR, Orientation_de_Gauche); Ecrire_Octet_EEPROM(PTR, Orientation_du_Centre);
     Ecrire_Octet_EEPROM(PTR, Orientation_de_Droite);
     for (byte I = 0; I < 12; I++) Ecrire_Octet_EEPROM(PTR, GROUPE_de_tete[I]);
     for (byte I = 0; I < 20; I++) Ecrire_Octet_EEPROM(PTR, FICHES_croisees[I]);} Serial.println();}

//====================== Traite le Rchargement d'une configuration  ======================

void Clone_Phases_avec_Orientations() {
   Phase_de_Gauche = Orientation_de_Gauche; Phase_du_Centre = Orientation_du_Centre;
   Phase_de_Droite = Orientation_de_Droite;}
   
void TFR_EEPROM_en_Initialisations() {
  PTR = 670; REFLECTEUR = Lire_Octet_EEPROM(PTR); PTR++;
  Rotor_de_Gauche = Lire_Octet_EEPROM(PTR); PTR++; Rotor_du_Centre = Lire_Octet_EEPROM(PTR); PTR++;
  Rotor_de_Droite = Lire_Octet_EEPROM(PTR); PTR++; Index_de_Gauche = Lire_Octet_EEPROM(PTR); PTR++;
  Index_du_Centre = Lire_Octet_EEPROM(PTR); PTR++; Index_de_Droite = Lire_Octet_EEPROM(PTR); PTR++;
  Orientation_de_Gauche = Lire_Octet_EEPROM(PTR); PTR++; Orientation_du_Centre = Lire_Octet_EEPROM(PTR);
  PTR++; Orientation_de_Droite = Lire_Octet_EEPROM(PTR); PTR++; Clone_Phases_avec_Orientations();
  for (byte I = 0; I < 12; I++) {GROUPE_de_tete[I] = Lire_Octet_EEPROM(PTR); PTR++;}
  for (byte I = 0; I < 20; I++) {FICHES_croisees[I] = Lire_Octet_EEPROM(PTR); PTR++;}}

void Restitue_Initialisation() {
  Serial.println(); Affiche_triple_fleche_ouvrantes(); Aff_TEXTE_EEPROM(155,10); Aff_TEXTE_EEPROM(128,16);
  Affiche_triple_fleche_fermantes(); Serial.println(); if (RESET) OUI = true; else Confirmer();
  if (OUI) {TFR_EEPROM_en_Initialisations(); Serial.println(); Affiche_la_CONFIGURATION_machine();}}

//================================= Traiter une COMMANDE ==================================

void Analyseur_syntaxique() {
  Num_ERR = 0; L_autorisee = 1; Caractere = Chaine_Memorisee[0]; if (Caractere == 'E') L_autorisee = 2;
  if (LGR_chaine > L_autorisee) Num_ERR = 1;
  if (Num_ERR == 0) {
     switch (Caractere) { // Refuse les caractères de commande invalides.
         case '?' : {break;} case 'I' : {break;} case 'S' : {break;} case 'R' : {break;} case 'M' : {break;}
         case 'V' : {break;} case 'T' : {break;} case 'L' : {break;} case 'C' : {break;} case 'P' : {break;}
         case 'E' : {break;} case '&' : {break;} case 'B' : {break;}
     default : Num_ERR = 2;}}
     if (Num_ERR > 0) Aff_ERREUR(Num_ERR);}

void Affiche_le_mode(byte Mode) {
  Affiche_passage_au_mode(); // " Passage en Mode "
  switch (Mode) {
    case 1 : {if (Mode_COMMANDE) {Aff_TEXTE_EEPROM(22,9); Serial.println('.');} // "Commande."
              else {Aff_TEXTE_EEPROM(375,9); Serial.println(); Serial.println();} break;} // "CRYPTAGE."
    case 2 : {Aff_TEXTE_EEPROM(253,7); Serial.println("(60 caracteres MAXI.)"); break;}
    case 3 : {Aff_TEXTE_EEPROM(266,8); Serial.println();}}} // "LETTRES."

void Traiter_la_COMMANDE() {
  byte I = 0; // Pointer le début de la chaine TEXTE.
  Analyseur_syntaxique(); // Vérifie les COMMANDES.
  if (Num_ERR == 0)
     switch (Caractere) {
       case '?' : {Affiche_le_MENU(); break;}
       case 'B' : {Bip = !Bip; Serial.println(); N_ESPACES(1); Aff_BIP(); Affiche_Egal();
                   Aff_OUI_NON(Bip); Serial.println(); break;} 
       case 'I' : {Mode_INITIALISATION(); break;}
       case 'E' : {Corriger_une_ERREUR(); break;}
       case 'S' : {Sauvegarde_Initialisation(); break;}
       case 'R' : {Restitue_Initialisation(); break;} // Ci-dessous " Morse sonore = "
       case 'M' : {Morse = !Morse; Aff_TEXTE_EEPROM(173,13); Affiche_Egal(); Aff_OUI_NON(Morse);
                   Serial.println(); break;}
       case 'V' : {Rapide = !Rapide;{ Aff_TEXTE_EEPROM(206,9); Aff_TEXTE_EEPROM(232,5);
                   Affiche_Egal();} // Ci-dessus " Vitesse Morse "
                   if (Rapide) {Aff_TEXTE_EEPROM(0,7); Serial.println();} // "Rapide."
                   else {Aff_TEXTE_EEPROM(7,6); Serial.println();} break;} // "Lente.")
       case 'T' : {Texte = true; Affiche_le_mode(2); break;}
       case 'L' : {Texte = false; Affiche_le_mode(3); break;}
       case 'C' : {Serial.println(); Affiche_la_CONFIGURATION_machine(); break;}
       case 'P' : {Affiche_espace_sous_la_PILE(); break;}
       // Ci_dessous : Replace les rotors en INITIALISATION avant de crypter un nouveau message.
       case '&' : {Mode_COMMANDE = false;
                   Installe_les_rotors_et_le_Reflecteur(); // Rotors Indexés, Orientés, et le bon Réflecteur.
                   Compteur_pour_separer = 0; Compteur_pour_ligne = 0; Serial.println();
                   Affiche_passage_au_mode(); Aff_CRYPTAGE(); Serial.println(); Serial.println();
                   Groupe_identification = true;}}}
   
  

