
/* Séquence qui permet D'exploiter les textes en EEPROM, ainsi que tester les cinq rotors et les
   deux réflecteurs préservés en mémoire non volatile EEPROM.
   Pour utiliser ce programme et vérifier les cinq rotors virtuel consulter l'exercice détaillé
   en page 25 du didacticiel.
   
            ************************************************************************
            * Ce démonstrateur purifié des séquences de test provisoires constitue *
            * le noyau du programme d'utilisation et sera complété au fur et à     *
            * mesure du développement.                                             *
            ************************************************************************
   
Compilateur v1.7.9 :                        PGM                 RAM dynamique
Avant modification :                  6598 Octets (21%)        1381 Octets (67%)
Passage des textes en EEPROM :        6036 Octets (19%)        517 Octets (25%)
Ajout 5 Rotors et 2 Réflecteurs :     6856 Octets (22%)        643 Octets (31%)   */

//--------------------- Définition des broches d'E/S ------------------------
#define BUZZER 4
#define LED_bleue 5
#define LED_Arduino 13

//------------------------ Constantes du programme --------------------------
#define Version "29/01/2024" // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Vitesse_USB 57600  // Taux de transfert sur la ligne USB @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Taille 60 // Longueur maximale de la chaine de dialogue @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// Taille réservée pour une chaine de 60 plus le délimiteur.
// Inutile de réserver plus de 60 caractère car la limite du FIFO USB est de 63 caractères.

//------------------------- Variables du programme --------------------------

boolean Direct; // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
byte Num_Rotor; // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

     // Rapide : Cadence de la manipulation Morse.
     // Texte : Mode TEXTE au lieu du mode LETTRES.
boolean Mode_COMMANDE, Morse, Rapide, Texte;

     // Caractere contiendra l'ordre alphabétique du caractère en cours de traitement.
     // Index : Pointeur pour le cryptage dans la chaine TEXTE.
byte Caractere, Compteur_pour_separer, Compteur_pour_ligne, ERREUR, Index, LGR_chaine, L_autorisee;

int PTR;

byte CODE_MORSE[27] = {0,66,129,133,97,32,132,99,128,64,142,101,130,67,65,
                       103,134,139,98,96,33,100,129,102,137,141,131};
     // NOTE : Pour simplifier le codage, l'indice 0 ne sera jamais utilisé et on aura
     //        Indice 1 pour A, indide 2 pour B etc.
     
byte Element_machine[26] = {};

unsigned long TOP_chronometre;

char Chaine_Memorisee[Taille+2]; // Tampon mémoire pour la chaine entrante. taille + 2 car il faut un
                          // emplacement pour le délimiteur et un emplacement pour "pouvoir déborder".

//-------------------------------------------------------------------------------------

void setup() {
  Serial.begin(57600); Serial.println(); Serial.print("Version du ");
  Serial.println(Version); Serial.println();
  pinMode(LED_Arduino, OUTPUT); pinMode(LED_bleue, OUTPUT); digitalWrite(LED_bleue,LOW);
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER,LOW);
  //--------------------------- Initialisation des variable ---------------------------
  L_autorisee = 1; Mode_COMMANDE = true; Morse = false; Rapide = false; Texte = false;
  TOP_chronometre = 0;
  Affiche_le_MENU();
  Num_Rotor = 1; Direct = true; Recharge_un_rotor(Num_Rotor);} //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 
void loop() {
  Attendre_une_chaine_sur_RX();
  if (Mode_COMMANDE) Traiter_la_COMMANDE(); else Traiter_le_CRYPTAGE();}

//============================== Routine provisoire ============================

// Dans "Faciliter_la_manipulation()" rétablir le passage de '1' à &. @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// Analyseur_syntaxique() : Enlever la valisation de 'D'. @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

void Recharge_un_rotor(byte Num_Rotor) {
  switch (Num_Rotor) {
    case 1 : {if (Direct) PTR = 712 ; else PTR = 738 ; break;} // Rotor I.
    case 2 : {if (Direct) PTR = 764 ; else PTR = 790 ; break;} // Rotor II
    case 3 : {if (Direct) PTR = 816 ; else PTR = 842 ; break;} // Rotor III
    case 4 : {if (Direct) PTR = 868 ; else PTR = 894 ; break;} // Rotor IV
    case 5 : {if (Direct) PTR = 920 ; else PTR = 946 ; break;} // Rotor V
    case 6 : {PTR = 972 ; break;} // Réflecteur B
    case 7 : PTR = 998 ;} // Réflecteur C
  for (byte I = 0; I < 26; I++) {Element_machine[I] = Lire_Octet_EEPROM(PTR); PTR++;}}

void Affiche_num_rotor() {Serial.println(); Serial.print("Num Rotor = ");
  Serial.println(Num_Rotor);}

void Affiche_Reflecteur() {Serial.println(); Serial.print("Reflecteur = ");
  if (Caractere == '6') Serial.println('B'); else Serial.println('C');}

void Affiche_sens_de_chiffrement() {Serial.println(); Serial.print("Sens chiffrement = ");
  if (Direct) Serial.println("Direct."); else Serial.println("Reflechi.");}

void Affiche_rechargement() {Serial.println(); 
  if (Num_Rotor < 6) {Serial.print("Charge Rotor "); Serial.print(Num_Rotor);
     Serial.print(" (Sens "); if (Direct) Serial.println("Direct.)"); else Serial.println("Reflechi.)");}
  else {Serial.print("Charge Reflecteur ");
        if (Num_Rotor == 6) Serial.println('B'); else Serial.println('C');}}

//==================== Procédures Écriture / Lecture EEPROM ====================
byte Lire_Octet_EEPROM(int PTR) {
  return eeprom_read_byte((unsigned char *) PTR);}
  
byte Ecrire_Octet_EEPROM(int PTR, byte OCTET) {
  eeprom_write_byte((unsigned char*) PTR,OCTET);}

int Lire_Entier_EEPROM(int PTR) {
  return eeprom_read_word((unsigned int *) PTR);}
  
void Ecrire_Entier_EEPROM(int PTR, int ENTIER) {
  eeprom_write_word((unsigned int *) PTR, ENTIER);}

void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {
    for (byte I=0; I<Longueur; I++) {Serial.print(char(Lire_Octet_EEPROM(PTR))); PTR++;};}

//======================== Procédures de servitude =============================

void BIP() {digitalWrite(BUZZER,HIGH); delay(150); digitalWrite(BUZZER,LOW);}

void N_ESPACES(byte Nb) {for(byte N = 0; N < Nb; N++)  Serial.print(' ');}

void Aff_46_etoiles() {for(byte N = 0; N < 46; N++) Serial.print('*');}

void Separer() {Aff_46_etoiles(); Serial.println(); Serial.print('*');}

void Terminer(byte Nb) {
  for(byte N = 0; N < Nb; N++) Serial.print(' '); Serial.println('*'); Serial.print('*');}

void Affiche_caractere_Num() {Serial.print(char(Caractere + 64));}

void Aff_OUI_NON(boolean OUI) {if (OUI) Aff_TEXTE_EEPROM(187,3); else Aff_TEXTE_EEPROM(193,3);}

void Afficher_la_Commande() {
  if (Mode_COMMANDE) {Aff_TEXTE_EEPROM(21,9); Aff_TEXTE_EEPROM(42,4);} // " Commande -> "
  else {if (!Texte) {Serial.print(' '); Aff_TEXTE_EEPROM(32,14);}}} // " Une lettre -> ")

void Affiche_le_MENU() {
  Serial.print('*'); Separer(); N_ESPACES(13); Aff_TEXTE_EEPROM(13,19); // "Menu des COMMANDES."
  ; Terminer(13);
  Separer(); Aff_TEXTE_EEPROM(45,31); // " ? ou , : Rappel des commandes."
  Terminer(14); Aff_TEXTE_EEPROM(76,30); // " I ou i : Mode INITIALISATION."
  Terminer(15); Aff_TEXTE_EEPROM(106,39); Terminer(6); // " S ou s : Sauvegarder la configuration."
  Aff_TEXTE_EEPROM(145,10); Aff_TEXTE_EEPROM(155,10); Aff_TEXTE_EEPROM(128,17);
  // Ci-dessus : " R ou r : Restituer la configuration."
  Terminer(8); Aff_TEXTE_EEPROM(164,33);  // " M ou m : Morse sonore OUI / NON."
  Terminer(12); Aff_TEXTE_EEPROM(197,41); // " V ou v : Vitesse du débit du code Morse."
  Terminer(4); Aff_TEXTE_EEPROM(238,21); Terminer(24); // " T ou t : Mode TEXTE."
  Aff_TEXTE_EEPROM(259,7); Aff_TEXTE_EEPROM(245,8); Aff_TEXTE_EEPROM(266,8);
  // Ci-dessus :" L ou l : Mode LETTRES."
  Terminer(22); Aff_TEXTE_EEPROM(274,35); // " C ou c : Affiche la CONFIGURATION."
  Terminer(10); Aff_TEXTE_EEPROM(309,41); // " P ou p : Espace disponible sous la PILE."
  Terminer(4); Aff_TEXTE_EEPROM(350,34); // " & ou 1 : Passer en Mode CRYPTAGE."
  Terminer(11); Serial.print(" G ou g : Groupe de t"); Serial.print(char(234)); Serial.print("te.");
  Terminer(20); Separer(); N_ESPACES(13); Aff_TEXTE_EEPROM(13,5);
  Aff_TEXTE_EEPROM(91,15); Terminer(12); Separer(); // "Menu INITIALISATION."
  Aff_TEXTE_EEPROM(384,43); Terminer(2); // " R ou r : Choix du ROTOR et de sa position ."
  Aff_TEXTE_EEPROM(427,40); Terminer(5); // " B ou b et C ou c : Choix du REFLECTEUR."
  Aff_TEXTE_EEPROM(467,41); Terminer(4);  // " F ou f : Configurer les FICHES croisées."
  Aff_TEXTE_EEPROM(508,35); Terminer(10); // " G ou g : Groupes d'identification."
  Aff_TEXTE_EEPROM(543,37);  Terminer(8); // " E ou e suivi de R, F ou G : ERREUR."
  Aff_46_etoiles(); Serial.println(); Serial.println();}

void POINT() {// Génère automatiquement le silence de séparation des éléments.
  if (Morse) {
      int T_element = 84; if (Rapide) T_element = 57;
      digitalWrite(LED_bleue,HIGH); digitalWrite(BUZZER,HIGH); delay(T_element);
      digitalWrite(LED_bleue,LOW); digitalWrite(BUZZER,LOW); delay(T_element);}
  Serial.print('.');}
  
void TRAIT() {// Génère automatiquement le silence de séparation des éléments.
  if (Morse) {
      int T_element = 252; if (Rapide) T_element = 171;
      digitalWrite(LED_bleue,HIGH); digitalWrite(BUZZER,HIGH); delay(T_element);
      digitalWrite(LED_bleue,LOW); digitalWrite(BUZZER,LOW); delay(T_element);}
  Serial.print('-');}

void TRANSCODE() {// Génère en Morse le Caractère représenté par l'ordre de la lettre dans l'alphabet.
 byte Lettre, Nb_Elements;
 int T_element = 252; if (Rapide) T_element = 171;
 Lettre = CODE_MORSE[Caractere]; Serial.print("  ");
 Nb_Elements = Lettre >> 5; // Isole le nombre d'éléments par cinq décalages à droite.
 for (Nb_Elements; Nb_Elements > 0; Nb_Elements--) {
   if ((Lettre & 1) > 0) TRAIT(); else POINT(); Lettre = Lettre >> 1;}
 delay (T_element); Serial.println();} // Silence de 0,4S avec le délai entre éléments.

//====================================================================================================

void Faciliter_la_manipulation() {
  switch(Caractere) {
    case ',' : {Caractere = '?'; break;} // case '1' : {Caractere = '&'; break;}
    case 224 : {Caractere = 'a'; break;} case 226 : {Caractere = 'a'; break;}
    case 232 : {Caractere = 'e'; break;} case 233 : {Caractere = 'e'; break;}
    case 234 : {Caractere = 'e'; break;} case 235 : {Caractere = 'e'; break;}
    case 238 : {Caractere = 'i'; break;} case 239 : {Caractere = 'i'; break;}
    case 244 : {Caractere = 'o'; break;} case 231 : {Caractere = 'c'; break;}
    case 251 : {Caractere = 'u'; break;} case 252 : {Caractere = 'u'; break;}
    case '.' : Caractere = ' ';}} // Le '.' sera ignoré comme l'espace.

void Attendre_une_chaine_sur_RX() {
  byte I = 0; LGR_chaine = 0; // Pointeur dans Chaine_Memorisee.
  Afficher_la_Commande();
  while(!Serial.available()) { // Attendre un premier caractere.
     //================= Chronométrage pour le délai écoulé ================
     if (millis() - TOP_chronometre > 50)  {
        TOP_chronometre = millis(); digitalWrite(LED_Arduino,!digitalRead(LED_Arduino));}}
     //---------------------------------------------------------------------
  while(Serial.available()) {
     Caractere = Serial.read();
     Faciliter_la_manipulation();
     if ((Caractere > 96) && (Caractere < 123)) Caractere = Caractere & 223; // Bien moins couteux
              // en octets que Chaine_Memorisee.toUpperCase() pour forcer la chaine en MAJUSCULES.
     if (LGR_chaine < (Taille + 1)) {
        Chaine_Memorisee[LGR_chaine] = Caractere; LGR_chaine++; delay(5);}} // Délai impératif.
  Chaine_Memorisee[LGR_chaine] = '\0'; // Placer le délimiteur de fin de texte.
  //----------------------------------------------------------------------------------------
  if (Mode_COMMANDE) {Serial.print('[');
     while(Chaine_Memorisee[I] != '\0') {Serial.print(char(Chaine_Memorisee[I])); I++;}
     Serial.print(']'); Serial.println(); digitalWrite(LED_Arduino, LOW);}}
  
void Mode_INITIALISATION() {Serial.println(); Aff_TEXTE_EEPROM(85,21); // " Mode INITIALISATION."
  Serial.println();} 

void Sauvegarde_Initialisation() {Serial.println(" Traite la SAUVEGARDE.");}

void Restitue_Initialisation() {Serial.println(" Traite le RECHARGEMENT.");}

void Affiche_la_CONFIGURATION() {Serial.println(" Affiche la CONFIGURATION.");}

int SRAM_LIBRE() { // Fonction qui retourne la taille de SRAM disponible.
  extern int __heap_start, *__brkval; // Déclaration des deux pointeurs dédiés.
  byte BIDON; // Dernière variable allouée, donc occupe le "haut" du TAS.
  if (__brkval == 0) {return (int) &BIDON -(int) &__heap_start;}
  else {return (int) &BIDON -(int) __brkval;}}

void Affiche_espace_sous_la_PILE() {
  Serial.print(" RAM DYNAMIQUE (PILE - TAS)  =  "); Serial.print (SRAM_LIBRE());
  Serial.println(" octets.");}

//================================== Traiter un CRYPTAGE ==================================

void Crypte_le_caractere() {// Reçoit l'ordre alphabétique du caractère crypté.
  Caractere = Element_machine[Caractere-1];} // Traverse le composant.
  
void Aff_texte_avant_erreur() {
     if (Texte) {Serial.println(); Serial.print('<');
         for(byte I = 0; I < Index; I++) Serial.print(Chaine_Memorisee[I]); Serial.print("> ");}}
 
void Aff_texte_Caractere() {Aff_TEXTE_EEPROM(602,10);} // " caractère"
void Aff_texte_invalide() {Aff_TEXTE_EEPROM(612,10);} // " invalide."

void Aff_ERREUR(byte Num_ERR) {
  switch (Num_ERR) {
     case 1 : { Serial.println(); Serial.print(' ');
                Aff_TEXTE_EEPROM(580,22); // " Attention : LGR > MAX."
                Index = 60; Aff_texte_avant_erreur(); Index = 0;
                L_autorisee = 255; // Pour éviter autant de BIPs que de dépassement.
                Compteur_pour_separer = 255; Compteur_pour_ligne = 255; break;}
     case 2 : {Aff_texte_Caractere(); Aff_TEXTE_EEPROM(411,4); Aff_TEXTE_EEPROM(22,8);
               Aff_texte_invalide(); break;} // Ci-dessus : " Caractère de COMMANDE invalide."
     case 3 : {if (Texte) {
                   Aff_texte_avant_erreur(); Index = LGR_chaine;} // Ne pas envoyer la suite.    
                   Serial.print('['); Serial.print(char(Caractere));
                   Serial.print(']'); Aff_texte_Caractere(); Aff_TEXTE_EEPROM(411,4);
                   Aff_TEXTE_EEPROM(375,8); Aff_texte_invalide();}}
                   // Ci-dessus : " Caractere de CRYPTAGE invalide."
  Serial.println(); BIP();}
  
void Verification_du_cryptage() {
  ERREUR = 0; Caractere = Chaine_Memorisee[Index];
  if (LGR_chaine > L_autorisee) ERREUR = 1;
  if ((Caractere < 'A') || (Caractere > 'Z')) ERREUR = 3;
  if ((Caractere == ' ') && (!Texte)) ERREUR = 3;
  if (ERREUR > 0) Aff_ERREUR(ERREUR);}

void Aff_retour_Commande() {Serial.println(); Aff_TEXTE_EEPROM(649,7); Aff_TEXTE_EEPROM(641,9);
     Aff_TEXTE_EEPROM(22,10); Serial.println();} // " Retour au mode COMMANDE."

void Traiter_le_Caractere() {
  if (Caractere == '&') Mode_COMMANDE = true;
  if (Mode_COMMANDE) Aff_retour_Commande();
  else {Verification_du_cryptage();
       if (ERREUR == 0) {
          Caractere = (byte(Chaine_Memorisee[Index])) - 64; Crypte_le_caractere();
          Affiche_caractere_Num();}}}

void Traiter_le_TEXTE() {// Index = 0 en entrée de cette procédure.
  for(Index; Index < LGR_chaine; Index++) {
      Caractere = Chaine_Memorisee[Index];
      if (!(Caractere == ' ')) {  // Ignorer ' ' en mode TEXTE.
         Traiter_le_Caractere(); Compteur_pour_separer++; Compteur_pour_ligne++;
      if (Compteur_pour_separer == 6) {Serial.print(' '); Compteur_pour_separer = 0;}
      if (Compteur_pour_ligne == 54) {Serial.println(); Compteur_pour_ligne = 0;}}}}

void Traiter_le_CRYPTAGE() { Index = 0; // Impératif pour Verification_du_cryptage().
  if (Texte) {L_autorisee = Taille; Traiter_le_TEXTE();}
  else {Caractere = Chaine_Memorisee[0]; L_autorisee = 1;
        if (Caractere != '&') Verification_du_cryptage();
        if (ERREUR == 0) {
           if (Caractere == '&') {
               Mode_COMMANDE = true; Serial.println(); Aff_retour_Commande();}
           else {Serial.print('['); Serial.print(char(Caractere)); 
                 Aff_TEXTE_EEPROM(622,11); // "] devient ["
                 Traiter_le_Caractere(); Serial.print(']'); TRANSCODE();}}
        else  ERREUR = 0;}} // Impératif pour reprendre la main avec '&'.

//================================= Traiter une COMMANDE ==================================

void Analyseur_syntaxique() { // Refuse les caractères de commande invalides.
  ERREUR = 0; L_autorisee = 1; Caractere = Chaine_Memorisee[0];
  if (LGR_chaine > L_autorisee) ERREUR = 1; 
  switch (Caractere) {
      case '?' : {; break;} case 'I' : {; break;} case 'S' : {; break;} case 'R' : {; break;}
      case 'M' : {; break;} case 'V' : {; break;} case 'T' : {; break;} case 'L' : {; break;}
      case 'C' : {; break;} case 'P' : {; break;} case '&' : {; break;} 
      case 'D' : {; break;} // A enkever @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@}
  default : ERREUR = 2;}
  if (ERREUR > 0) Aff_ERREUR(ERREUR);}

void Affiche_passage_au_mode() {Aff_TEXTE_EEPROM(633,17);}

void Affiche_le_mode(byte Mode) {
  Affiche_passage_au_mode(); // " Passage en Mode "
  switch (Mode) {
    case 1 : {if (Mode_COMMANDE) {Aff_TEXTE_EEPROM(22,9); Serial.println('.');} // "Commande."
              else {Aff_TEXTE_EEPROM(375,9); BIP(); Serial.println();} break;} // "CRYPTAGE."
    case 2 : {Aff_TEXTE_EEPROM(253,7); Serial.println("(60 caracteres MAXI.)"); break;}
    case 3 : {Aff_TEXTE_EEPROM(266,8); Serial.println();}}} // "LETTRES."

void Traiter_la_COMMANDE() {
  byte I = 0; // Pointer le début de la chaine TEXTE.
  if ((Caractere < '1') || (Caractere > '7')) // A enlever. @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  Analyseur_syntaxique(); // Vérifie les COMMANDES.
  if (ERREUR == 0)
     switch (Caractere) {
       case '1' : {Num_Rotor = 1; Affiche_num_rotor(); break;}
       case '2' : {Num_Rotor = 2; Affiche_num_rotor(); break;}
       case '3' : {Num_Rotor = 3; Affiche_num_rotor(); break;}
       case '4' : {Num_Rotor = 4; Affiche_num_rotor(); break;}
       case '5' : {Num_Rotor = 5; Affiche_num_rotor(); break;}
       case '6' : {Num_Rotor = 6; Affiche_Reflecteur(); break;} // Reflecteur 'B';
       case '7' : {Num_Rotor = 7; Affiche_Reflecteur(); break;} // Reflecteur 'C';
       case 'D' : {Direct = true; Affiche_sens_de_chiffrement(); break;}
       case 'R' : {Direct = false; Affiche_sens_de_chiffrement(); break;}      
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
       case '?' : {Affiche_le_MENU(); break;}
       case 'I' : {Affiche_rechargement(); Recharge_un_rotor(Num_Rotor); break;}
//       case 'I' : {Mode_INITIALISATION(); break;} A rétablir @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
       case 'S' : {Sauvegarde_Initialisation(); break;}
//       case 'R' : {Restitue_Initialisation(); break;} // Ci-dessous " Morse sonore = "
       case 'M' : {Morse = !Morse; Aff_TEXTE_EEPROM(173,14); Serial.print("= "); Aff_OUI_NON(Morse);
                   Serial.println(); break;}
       case 'V' : {Rapide = !Rapide;{ Aff_TEXTE_EEPROM(206,9); Aff_TEXTE_EEPROM(232,5);
                   Serial.print(" = ");} // Ci-dessus " Vitesse Morse "
                   if (Rapide) {Aff_TEXTE_EEPROM(0,7); Serial.println();} // "Rapide."
                   else {Aff_TEXTE_EEPROM(7,6); Serial.println();} break;} // "Lente.")
       case 'T' : {Texte = true; Affiche_le_mode(2); break;}
       case 'L' : {Texte = false; Affiche_le_mode(3); break;}
       case 'C' : {Affiche_la_CONFIGURATION(); break;}
       case 'P' : {Affiche_espace_sous_la_PILE(); break;}
       case '&' : {Mode_COMMANDE = false; Compteur_pour_separer = 0; Compteur_pour_ligne = 0;
                   Serial.println(); Affiche_passage_au_mode(); Aff_TEXTE_EEPROM(375,9);
                   Serial.println();}}}  // " Passage au mode CRYPTAGE.");
   
  

